import scipy.optimize
from prophetverse.budget_optimization.base import (
    BaseBudgetOptimizer,
    BaseConstraint,
    BaseOptimizationObjective,
    BaseParametrizationTransformation,
)
import scipy
import warnings
import numpy as np
from jax import grad, hessian
import jax.numpy as jnp
import jax
import pandas as pd
from prophetverse.budget_optimization.parametrization_transformations import (
    IdentityTransform,
)
from typing import List, Optional
import numpyro

__all__ = ["BudgetOptimizer"]


class BudgetOptimizer(BaseBudgetOptimizer):
    """Budget optimizer using scipy.optimize.minimize.

    Parameters
    ----------
    objective : BaseOptimizationObjective
        Objective function object
    constraints : list of BaseConstraint
        List of constraint objects
    decision_variable_transform : BaseDecisionVariableTransform, optional
        Decision variable transform object
    method : str, optional
        Optimization method to use. Default is "SLSQP".
    tol : float, optional
        Tolerance for termination. Default is None.
    bounds : Union[List[tuple], dict[str, tuple]], optional
        Bounds for decision variables. If a list, the value is used directly
        in scipy.optimize.minimize. If a dict, the keys are the column names and
        the values are the bounds for each column. Default is (0, np.inf) for
        each column.
    options : dict, optional
        Options for the optimization method. Default is None.
    callback : callable, optional
        Callback function to be called after each iteration. Default is None.
    """

    def __init__(
        self,
        objective: BaseOptimizationObjective,
        constraints: List[BaseConstraint],
        parametrization_transform: Optional[BaseParametrizationTransformation] = None,
        method="SLSQP",
        tol=None,
        bounds=None,
        options=None,
        callback=None,
    ):

        self.method = method
        self.options = options
        self.bounds = bounds
        self.tol = tol
        self.parametrization_transform = parametrization_transform
        self.callback = callback
        super().__init__(constraints=constraints, objective=objective)

    def _optimize(self, model, X, horizon, columns):

        # To avoid errors with Scipy optimize
        numpyro.enable_x64()
        self.set_predictive_attr(
            model=model,
            X=X,
            horizon=horizon,
            columns=columns,
        )

        # Decision variable transform
        self._parametrization_transform = self.parametrization_transform
        if self.parametrization_transform is None:
            self._parametrization_transform = IdentityTransform()

        # Prepare bounds
        self._bounds = self.bounds
        if self._bounds is None:
            self._bounds = {col: (0, np.inf) for col in columns}

        # Set constraints
        self.constraints_ = []
        for constraint in self.constraints:
            self.constraints_.append(
                constraint(
                    X=X,
                    horizon=horizon,
                    columns=columns,
                )
            )

        self.objective_fun_ = self.objective(
            model=model,
            X=X,
            horizon=horizon,
            columns=columns,
        )

        x0 = X.loc[X.index.get_level_values(-1).isin(horizon), columns].values.flatten()
        x0 = x0.astype(jnp.float64)
        # Transform decision variable
        self._parametrization_transform.fit(X, horizon, columns)
        x0 = self._parametrization_transform.transform(x0)

        self.x0_ = x0

        # Bounds are set based on the re-parametrized decision variable.
        self.bounds_ = []
        if isinstance(self._bounds, list):
            self.bounds_ = self._bounds
        elif len(x0) % len(columns) == 0:
            size_per_column = len(x0) // len(columns)
            for col in columns:
                self.bounds_.extend(
                    [self._bounds.get(col, (0, np.inf))] * size_per_column
                )
        else:
            if self._bounds is not None:
                warnings.warn(
                    "Bounds are not set correctly. Using default bounds (0, np.inf) for each decision variable."
                )
            self.bounds_ = [(0, np.inf)] * len(x0)

        self.objective_fun_ = self.wrap_func_with_inv_transform(self.objective_fun_)
        self.jac_ = grad(self.objective_fun_)
        self.hess_ = hessian(self.objective_fun_)

        for i in range(len(self.constraints_)):
            fun = self.constraints_[i]["fun"]
            self.constraints_[i]["fun"] = self.wrap_func_with_inv_transform(fun)
            self.constraints_[i]["jac"] = grad(self.constraints_[i]["fun"])
            self.constraints_[i]["args"] = (self,)

        minimize_kwargs = dict(
            fun=self.objective_fun_,
            x0=x0,
            jac=self.jac_,
            method=self.method,
            constraints=self.constraints_,
            hess=self.hess_,
            bounds=self.bounds_,
            options=self.options,
            tol=self.tol,
            args=self,
            callback=self.callback,
        )

        if self.method in ["BFGS"]:
            # Avoid warnings about bounds and constraints
            # but warn if the user tries to use method that do not use them.
            if self.bounds is not None:
                warnings.warn(
                    f"{self.method} method does not support bounds. Ignoring bounds."
                )
            if self.constraints_:
                warnings.warn(
                    f"{self.method} method does not support constraints. Ignoring constraints."
                )

            minimize_kwargs.pop("constraints")
            minimize_kwargs.pop("bounds")
            minimize_kwargs.pop("hess")

        res = scipy.optimize.minimize(**minimize_kwargs)
        self.result_ = res

        X_opt = X.copy()
        x_opt = self._parametrization_transform.inverse_transform(res.x)
        mask = X_opt.index.get_level_values(-1).isin(horizon)
        X_opt.loc[mask, columns] = x_opt.reshape(-1, len(columns))
        return X_opt

    def set_predictive_attr(self, model, X, horizon, columns):

        fh: pd.Index = X.index.get_level_values(-1).unique()
        # Get the indexes of `horizon` in fh
        horizon_idx = jnp.array([fh.get_loc(h) for h in horizon])

        predictive = model.optimize_predictive_callable(
            X=X, horizon=horizon, columns=columns
        )
        self.predictive_no_jit_ = predictive
        self.predictive_ = jax.jit(predictive)
        self.horizon_idx_ = horizon_idx

    def wrap_func_with_inv_transform(self, fun):
        """Wrap a function with parametrization inverse transform"""

        def wrapper(x, *args):
            xt = self._parametrization_transform.inverse_transform(x)
            return fun(xt, *args)

        return wrapper

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        from prophetverse.budget_optimization.objectives import (
            MaximizeROI,
            MaximizeKPI,
            MinimizeBudget,
        )
        from prophetverse.budget_optimization.constraints import (
            TotalBudgetConstraint,
            MinimumTargetResponse,
        )
        from prophetverse.budget_optimization.parametrization_transformations import (
            InvestmentPerChannelTransform,
            TotalInvestmentTransform,
            InvestmentPerChannelAndSeries,
            InvestmentPerSeries,
        )

        params = []

        for parametrization in [
            None,
            InvestmentPerChannelTransform(),
            TotalInvestmentTransform(),
            InvestmentPerChannelAndSeries(),
            InvestmentPerSeries(),
        ]:

            if not isinstance(
                parametrization, (TotalInvestmentTransform, InvestmentPerSeries)
            ):
                params.extend(
                    [
                        {
                            "objective": MaximizeKPI(),
                            "constraints": [
                                TotalBudgetConstraint(),
                            ],
                            "parametrization_transform": parametrization,
                            "options": {"maxiter": 10},
                        },
                        {
                            "objective": MaximizeROI(),
                            "constraints": [
                                TotalBudgetConstraint(),
                                MinimumTargetResponse(0.5),
                            ],
                            "options": {"maxiter": 10},
                        },
                    ]
                )

            params.extend(
                [
                    {
                        "objective": MinimizeBudget(),
                        "constraints": [
                            MinimumTargetResponse(0.5),
                        ],
                        "options": {"maxiter": 10},
                    },
                ]
            )
        return params



"""Numpyro inference engines for prophet models.

The classes in this module take a model, the data and perform inference using Numpyro.
"""

from typing import Optional

import jax.numpy as jnp
import numpyro
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.autoguide import AutoDelta
from numpyro.infer.initialization import init_to_mean
from numpyro.infer.svi import SVIRunResult

from prophetverse.engine.base import BaseInferenceEngine
from prophetverse.engine.optimizer.optimizer import (
    AdamOptimizer,
    BaseOptimizer,
    LBFGSSolver,
)
from prophetverse.utils.deprecation import deprecation_warning

_DEFAULT_PREDICT_NUM_SAMPLES = 1000
DEFAULT_PROGRESS_BAR = False


class MAPInferenceEngine(BaseInferenceEngine):
    """
    Maximum a Posteriori (MAP) Inference Engine.

    This class performs MAP inference using Stochastic Variational Inference (SVI)
    with AutoDelta guide. It provides methods for inference and prediction.

    Parameters
    ----------
    optimizer : Optional[BaseOptimizer]
        The optimizer to be used for inference. If not provided, the default is
        LBFGSSolver.
    num_steps : int, optional
        The number of steps to run the optimizer. Default is 10000.
    num_samples : int, optional
        The number of samples to generate during prediction.
        Default is _DEFAULT_PREDICT_NUM_SAMPLES.
    rng_key : optional
        The random number generator key.
    progress_bar : bool, optional
        Whether to display a progress bar during inference. Default is DEFAULT_PROGRESS_BAR.
    stable_update : bool, optional
        Whether to use stable update during inference. Default is False.
    forward_mode_differentiation : bool, optional
        Whether to use forward mode differentiation. Default is False.
    init_loc_fn : optional
        The function to initialize the location parameter. If not provided, the default is init_to_mean.

    """

    _tags = {
        "inference_method": "map",
    }

    def __init__(
        self,
        optimizer: Optional[BaseOptimizer] = None,
        num_steps=10_000,
        num_samples=_DEFAULT_PREDICT_NUM_SAMPLES,
        rng_key=None,
        progress_bar: bool = DEFAULT_PROGRESS_BAR,
        stable_update=False,
        forward_mode_differentiation=False,
        init_loc_fn=None,
    ):

        self.optimizer = optimizer
        self.num_steps = num_steps
        self.num_samples = num_samples
        self.progress_bar = progress_bar
        self.stable_update = stable_update
        self.forward_mode_differentiation = forward_mode_differentiation
        self.init_loc_fn = init_loc_fn
        super().__init__(rng_key)

        if optimizer is None:
            optimizer = LBFGSSolver()

        self._optimizer = optimizer

        self._init_loc_fn = init_loc_fn
        if init_loc_fn is None:
            self._init_loc_fn = init_to_mean()

        self._num_steps = num_steps

        if self._optimizer.get_tag("is_solver", False):  # type: ignore[union-attr]
            # If solver, there's a single "solver step". For compatibility,
            # we set num_steps to 1 and max_iter to the original num_steps.

            self._optimizer = self._optimizer.set_max_iter(  # type: ignore[union-attr]
                self._num_steps
            )
            self._num_steps = 1

    def _infer(self, **kwargs):
        """
        Perform MAP inference.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        self
            The updated MAPInferenceEngine object.
        """
        self.guide_ = AutoDelta(self.model_, init_loc_fn=self._init_loc_fn)

        def get_result(
            rng_key,
            model,
            guide,
            optimizer,
            num_steps,
            progress_bar,
            stable_update,
            forward_mode_differentiation,
            **kwargs,
        ) -> SVIRunResult:
            svi_ = SVI(
                model,
                guide,
                optimizer,
                loss=Trace_ELBO(),
            )
            return svi_.run(
                rng_key=rng_key,
                progress_bar=progress_bar,
                stable_update=stable_update,
                num_steps=num_steps,
                forward_mode_differentiation=forward_mode_differentiation,
                **kwargs,
            )

        self.run_results_: SVIRunResult = get_result(
            self._rng_key,
            self.model_,
            self.guide_,
            self._optimizer.create_optimizer(),
            self._num_steps,
            stable_update=self.stable_update,
            progress_bar=self.progress_bar,
            forward_mode_differentiation=self.forward_mode_differentiation,
            **kwargs,
        )

        self.raise_error_if_nan_loss(self.run_results_)

        self.posterior_samples_ = self.guide_.sample_posterior(
            self._rng_key, params=self.run_results_.params, **kwargs
        )
        return self

    def raise_error_if_nan_loss(self, run_results: SVIRunResult):
        """
        Raise an error if the loss is NaN.

        Parameters
        ----------
        run_results : SVIRunResult
            The result of the SVI run.

        Raises
        ------
        MAPInferenceEngineError
            If the last loss is NaN.
        """
        losses = run_results.losses
        if jnp.isnan(losses)[-1]:
            msg = "NaN losses in MAPInferenceEngine."
            msg += " Try decreasing the learning rate or changing the model specs."
            msg += " If the problem persists, please open an issue at"
            msg += " https://github.com/felipeangelimvieira/prophetverse"
            raise MAPInferenceEngineError(msg)

    def _predict(self, **kwargs):
        """
        Generate predictions using the trained model.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        dict
            The predicted samples generated by the model.
        """
        predictive = numpyro.infer.Predictive(
            self.model_,
            params=self.run_results_.params,
            guide=self.guide_,
            # posterior_samples=self.posterior_samples_,
            num_samples=self.num_samples,
        )
        self.samples_ = predictive(rng_key=self._rng_key, **kwargs)
        return self.samples_

    @classmethod
    def get_test_params(*args, **kwargs):
        """Return test params for unit testing."""
        return [
            {
                "optimizer": LBFGSSolver(),
                "num_steps": 100,
            },
            {
                "optimizer": AdamOptimizer(),
                "num_steps": 100,
            },
        ]


class MAPInferenceEngineError(Exception):
    """Exception raised for NaN losses in MAPInferenceEngine."""

    def __init__(self, message="NaN losses in MAPInferenceEngine"):
        self.message = message
        super().__init__(self.message)



import numpyro
from prophetverse.effects.base import BaseEffect
from typing import Any, Dict, Optional
import jax.numpy as jnp
from prophetverse.utils.numpyro import CacheMessenger


def wrap_with_cache_messenger(model):

    def wrapped(*args, **kwargs):
        with CacheMessenger():
            return model(*args, **kwargs)

    return wrapped


def model(
    y,
    trend_model: BaseEffect,
    trend_data: Dict[str, jnp.ndarray],
    target_model: BaseEffect,
    target_data: Dict[str, jnp.ndarray],
    data: Optional[Dict[str, jnp.ndarray]] = None,
    exogenous_effects: Optional[Dict[str, BaseEffect]] = None,
    **kwargs,
):
    """
    Define the Prophet-like model for univariate timeseries.

    Parameters
    ----------
        y (jnp.ndarray): Array of time series data.
        trend_model (BaseEffect): Trend model.
        trend_data (dict): Dictionary containing the data needed for the trend model.
        data (dict): Dictionary containing the exogenous data.
        exogenous_effects (dict): Dictionary containing the exogenous effects.
        noise_scale (float): Noise scale.
    """

    with CacheMessenger():
        predicted_effects: Dict[str, jnp.ndarray] = {}

        with numpyro.handlers.scope(prefix="trend"):
            trend = trend_model(data=trend_data, predicted_effects=predicted_effects)

        predicted_effects["trend"] = numpyro.deterministic("trend", trend)

        # Exogenous effects
        if exogenous_effects is not None:
            for exog_effect_name, exog_effect in exogenous_effects.items():
                transformed_data = data[exog_effect_name]  # type: ignore[index]

                with numpyro.handlers.scope(prefix=exog_effect_name):
                    effect = exog_effect(transformed_data, predicted_effects)

                effect = numpyro.deterministic(exog_effect_name, effect)
                predicted_effects[exog_effect_name] = effect

        target_model.predict(target_data, predicted_effects)



import numpy as np
import pandas as pd

__all__ = ["load_synthetic_squared_exogenous"]


def _generate_dataset(
    n_periods: int,
    seasonality_period: int,
    trend_slope: float,
    exogenous_range: tuple,
    noise_std: float = 1.0,
    seed: int = 0,
) -> tuple[pd.Series, pd.DataFrame]:
    """
    Generate a simple synthetic time series in sktime format.

    The series is composed of seasonality, trend,
    and exogenous variables.

    Parameters
    ----------
    n_periods : int
        Number of time periods to simulate.
    seasonality_period : int
        Period of the seasonal component.
    trend_slope : float
        Slope of the linear trend.
    exogenous_range : tuple
        Range (min, max) for the exogenous variable values.
    noise_std : float, optional
        Standard deviation of the Gaussian noise, by default 1.0.
    seed : int, optional
        Random seed for reproducibility, by default None.

    Returns
    -------
    tuple[pd.Series, pd.DataFrame]
        y : pd.Series
            Target variable in sktime format with time index.
        X : pd.DataFrame
            Exogenous variables and components in sktime format with time index.

    Examples
    --------
    >>> y, X = generate_sktime_time_series(100, 12, 0.5, (1, 10), 0.5, seed=42)
    >>> y.head()
    time
    0    0.838422
    1    1.488498
    2    2.230748
    3    2.930336
    4    3.724452
    Name: target, dtype: float64
    >>> X.head()
       seasonality     trend  exogenous     noise
    time
    0      0.000000  0.000000   5.749081  0.211731
    1      0.258819  0.500000   6.901429  0.326080
    2      0.500000  1.000000   6.463987  0.460959
    3      0.707107  1.500000   5.197317  0.676962
    4      0.866025  2.000000   3.312037  0.546416
    """
    rng = np.random.default_rng(seed)

    # Time index
    time_index = pd.period_range(
        start="2010-01-01", freq="D", periods=n_periods, name="time"
    )

    # Seasonal component
    seasonality = np.sin(2 * np.pi * np.arange(n_periods) / seasonality_period)

    _t = np.arange(n_periods)
    _t = _t - _t.mean()
    _t = _t / n_periods * 20
    # Linear trend
    trend = trend_slope / (1 + np.exp(-_t))

    # Exogenous variable
    exogenous = rng.uniform(*exogenous_range, size=n_periods)

    # Logarithmic effect of exogenous variable
    exog_effect = 2 * (exogenous - 5) ** 2  # Adding 1 to avoid log(0)

    # Noise
    noise = rng.normal(scale=noise_std, size=n_periods)

    # Target variable
    target = seasonality + trend + exog_effect + noise

    # Construct y and X
    y = pd.Series(data=target, index=time_index, name="target").to_frame()
    X = pd.DataFrame(
        data={
            "exogenous": exogenous,
        },
        index=time_index,
    )

    return y, X


def load_synthetic_squared_exogenous():
    """Load the synthetic log exogenous dataset.

    This dataset is just for documentation purposes.

    Returns
    -------
    pd.DataFrame
        The synthetic target variable
    pd.DataFrame
        The synthetic exogenous variable

    """
    return _generate_dataset(
        n_periods=700,
        seasonality_period=365.25,
        trend_slope=10,
        exogenous_range=(1, 10),
        noise_std=2,
        seed=42,
    )



from prophetverse.budget_optimization.base import (
    BaseParametrizationTransformation,
)
import jax.numpy as jnp
from prophetverse.utils.frame_to_array import series_to_tensor

__all__ = [
    "IdentityTransform",
    "InvestmentPerChannelTransform",
    "TotalInvestmentTransform",
    "InvestmentPerChannelAndSeries",
    "InvestmentPerSeries",
]


class IdentityTransform(BaseParametrizationTransformation):
    """Identity decision variable transform.

    This transform does not change the decision variables. It is used when
    the decision variables are already in the correct format.
    """

    _tags = {
        "name": "identity",
        "backend": "scipy",
    }

    def _transform(self, x):
        return x

    def _inverse_transform(self, xt):
        return xt


class InvestmentPerChannelTransform(BaseParametrizationTransformation):
    """
    Change parametrization to be the share of each channel.

    Instead of parametrizing every time x channel investment, this transform
    parametrize to optimize per channel investment, while keeping the initial
    guess temporal share of the investment.
    """

    def _fit(self, X, horizon, columns):

        mask = X.index.get_level_values(-1).isin(horizon)
        X = X.loc[mask, columns]

        x_array = series_to_tensor(X)

        self.n_series_ = x_array.shape[0]
        # get daily share
        self.daily_share_ = x_array / x_array.sum(axis=(0, 1), keepdims=True)

    def _transform(self, x: jnp.ndarray):
        # get share per column
        # x is a (N*M) array
        x = x.reshape((self.n_series_, -1, len(self.columns_)))
        # get the sum of each row
        x_sum = jnp.sum(x, axis=(0, 1)).flatten()
        # get the share of each column
        return x_sum

    def _inverse_transform(self, xt):
        # Multiply each column share by the daily share
        xt = xt.reshape((1, 1, len(self.columns_)))

        xt = xt * self.daily_share_
        xt = xt.flatten()
        return xt


class TotalInvestmentTransform(BaseParametrizationTransformation):
    """
    Change parametrization to be the total investment.

    Instead of parametrizing every time x channel investment, this transform
    parametrize to optimize per channel investment, while keeping the initial
    guess temporal share of the investment.
    """

    def _fit(self, X, horizon, columns):
        mask = X.index.get_level_values(-1).isin(horizon)
        X = X.loc[mask, columns]

        x_array = series_to_tensor(X)

        # get daily share
        self.daily_share_ = x_array / x_array.sum()

    def _transform(self, x: jnp.ndarray):
        # get share per column
        # x is a (N*M) array
        x = x.reshape((-1, len(self.columns_)))
        # get the sum of each row
        x_sum = jnp.sum(x)
        # get the share of each column
        # Make sure x_sum is one dimensional
        x_sum = x_sum.reshape(-1)
        return x_sum

    def _inverse_transform(self, xt):
        # Multiply each column share by the daily share

        xt = xt * self.daily_share_
        xt = xt.flatten()
        return xt


class InvestmentPerChannelAndSeries(BaseParametrizationTransformation):

    def _fit(self, X, horizon, columns):

        mask = X.index.get_level_values(-1).isin(horizon)
        X = X.loc[mask, columns]

        x_array = series_to_tensor(X)

        self.n_series_ = x_array.shape[0]
        # get daily share
        self.daily_share_ = x_array / x_array.sum(axis=1, keepdims=True)

    def _transform(self, x: jnp.ndarray):
        # get share per column
        # x is a (N*M) array
        x = x.reshape((self.n_series_, -1, len(self.columns_)))
        # get the sum of each row
        x_sum = jnp.sum(x, axis=1).flatten()

        return x_sum

    def _inverse_transform(self, xt):
        # Multiply each column share by the daily share
        xt = xt.reshape((self.n_series_, 1, len(self.columns_)))

        xt = xt * self.daily_share_
        xt = xt.flatten()
        return xt


class InvestmentPerSeries(BaseParametrizationTransformation):

    def _fit(self, X, horizon, columns):

        mask = X.index.get_level_values(-1).isin(horizon)
        X = X.loc[mask, columns]

        x_array = series_to_tensor(X)

        self.n_series_ = x_array.shape[0]
        # get daily share
        self.daily_share_ = x_array / x_array.sum(axis=(1, 2), keepdims=True)

    def _transform(self, x: jnp.ndarray):
        # get share per column
        # x is a (N*M) array
        x = x.reshape((self.n_series_, -1, len(self.columns_)))
        # get the sum of each row
        x_sum = jnp.sum(x, axis=(1, 2)).flatten()

        return x_sum

    def _inverse_transform(self, xt):
        # Multiply each column share by the daily share
        xt = xt.reshape((self.n_series_, 1, 1))

        xt = xt * self.daily_share_
        xt = xt.flatten()
        return xt



from typing import Optional
import pandas as pd
from sktime.transformations.base import BaseTransformer
from sklearn.utils.validation import check_is_fitted


class EventsDummyTransformer(BaseTransformer):
    """
    Create holiday dummy variables from an events dataframe in fbprophet format.

    The expected events dataframe must have the following columns:
        - event_name: the name of the holiday event.
        - ds: the date of the event (datetime convertible).
        - lower_window: integer offset (in days) indicating how many days
          before the event the effect starts.
        - upper_window: integer offset (in days) indicating how many days
          after the event the effect ends.

    The transformer produces dummy variables such that for each event row the
    effective dates are all dates in the range ds + lower_window to ds + upper_window.
    If dummy_by_window is False, a single dummy column is created per event (with an
    optional prefix). If True, a separate dummy is created for each offset within the
    window (with the column name reflecting the offset).

    During transform, if X is None, the stored dummy dataframe (with a daily date
    range spanning from the earliest to the latest effective holiday date) is returned.
    Otherwise, a left join is performed on X so that only the dates available in X appear,
    while preserving the index dtype of X.

    Parameters
    ----------
    events_df : pd.DataFrame
        DataFrame of holiday events with columns: event_name, ds, lower_window, upper_window.
    prefix : Optional[str], default=None
        Optional prefix to add to the dummy column names.
    dummy_by_window : bool, default=False
        If True, creates a separate dummy for each window offset (each day in the holiday window).
        Otherwise, creates one dummy column per event.

    Attributes
    ----------
    events_df_ : pd.DataFrame
        Copy of events_df with ds converted to datetime.
    _dummy_df_ : pd.DataFrame
        Dummy dataframe covering the full date range from the earliest to the latest effective date.
    """

    _tags = {
        "fit_is_empty": False,
        "requires_X": False,
        "remember_data": False,
        "X_inner_mtype": "pd.DataFrame",
        "transform-returns-same-time-index": True,
    }

    def __init__(
        self,
        events_df: pd.DataFrame,
        prefix: Optional[str] = None,
        dummy_by_window: bool = False,
    ):
        self.events_df = events_df
        self.prefix = prefix
        self.dummy_by_window = dummy_by_window
        super().__init__()

    def _fit(self, X, y=None):
        """
        Fit the transformer.

        Converts the 'ds' column to datetime and precomputes the dummy DataFrame,
        storing it in self._dummy_df_.

        Parameters
        ----------
        X : pd.DataFrame or None
            Ignored.
        y : Ignored.

        Returns
        -------
        self : reference to self.
        """
        self.events_df_ = self.events_df.copy()
        self.events_df_["ds"] = pd.to_datetime(self.events_df_["ds"])

        # Build the dummy data from events.
        rows = []
        for _, row in self.events_df_.iterrows():
            event = row["event_name"]
            ds = row["ds"]
            lower = int(row["lower_window"])
            upper = int(row["upper_window"])
            if self.dummy_by_window:
                for offset in range(lower, upper + 1):
                    effective_date = ds + pd.Timedelta(days=offset)
                    col = (
                        f"{self.prefix}_{event}_offset_{offset}"
                        if self.prefix
                        else f"{event}_offset_{offset}"
                    )
                    rows.append((effective_date, col))
            else:
                col = f"{self.prefix}_{event}" if self.prefix else f"{event}"
                for offset in range(lower, upper + 1):
                    effective_date = ds + pd.Timedelta(days=offset)
                    rows.append((effective_date, col))
        if not rows:
            raise ValueError("No events found to transform.")

        temp_df = pd.DataFrame(rows, columns=["date", "column"])
        temp_df["value"] = 1
        dummy_df = temp_df.pivot_table(
            index="date", columns="column", values="value", aggfunc="max", fill_value=0
        )
        # Ensure dummy_df has a DatetimeIndex.
        dummy_df.index = pd.to_datetime(dummy_df.index)
        # Reindex to cover the full range.
        full_index = pd.date_range(
            start=dummy_df.index.min(), end=dummy_df.index.max(), freq="D"
        )
        self._dummy_df_ = dummy_df.reindex(full_index, fill_value=0)
        return self

    def _transform(self, X, y=None):
        """
        Transform the events into holiday dummy variables.

        If X is None, returns the stored dummy DataFrame.
        Otherwise, it left joins X with the dummy DataFrame so that only the dates
        in X are kept. The output index is converted back to the same type as X.index.

        Parameters
        ----------
        X : pd.DataFrame or None
            DataFrame whose index is used to subset the dummy DataFrame.
        y : Ignored.

        Returns
        -------
        X_transformed : pd.DataFrame
            Left-joined DataFrame of X and dummy variables.
        """
        check_is_fitted(self, attributes=["_dummy_df_"])

        dummy_df = self._dummy_df_.copy()
        if X is None:
            return dummy_df
        else:
            # Preserve the original index type of X.
            # First, determine an "output index" from X.index.
            if isinstance(X.index, pd.PeriodIndex):
                # Convert PeriodIndex to timestamp for reindexing dummy_df.
                ts_index = X.index.to_timestamp()
                dummy_reindexed = dummy_df.reindex(ts_index, fill_value=0)
                # Convert back to PeriodIndex with same frequency.
                out_index = pd.PeriodIndex(dummy_reindexed.index, freq=X.index.freq)
                dummy_reindexed.index = out_index
            elif isinstance(X.index, (pd.DatetimeIndex, pd.RangeIndex)):
                ts_index = pd.to_datetime(X.index)
                dummy_reindexed = dummy_df.reindex(ts_index, fill_value=0)
                out_index = dummy_reindexed.index
            else:
                # For any other index type, attempt conversion to DatetimeIndex.
                ts_index = pd.to_datetime(X.index, errors="coerce")
                if ts_index.isnull().any():
                    raise ValueError(
                        "X.index could not be converted to a DatetimeIndex."
                    )
                dummy_reindexed = dummy_df.reindex(ts_index, fill_value=0)
                out_index = dummy_reindexed.index

            # Now, perform a left join: we keep X's columns and add dummy columns.
            X_copy = X.copy()
            X_copy = X_copy.reindex(out_index)
            X_joined = X_copy.join(dummy_reindexed, how="left")
            X_joined.fillna(0, inplace=True)
            # Ensure the output index has the same dtype as the input.
            X_joined.index = X.index
            return X_joined

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """
        Return testing parameter settings for the EventsDummyTransformer.

        Returns
        -------
        params : list of dict
            A list of parameter dictionaries.
        """
        params = []
        test_events = pd.DataFrame(
            {
                "event_name": ["holiday1", "holiday2"],
                "ds": [pd.Timestamp("2020-01-01"), pd.Timestamp("2020-01-10")],
                "lower_window": [-1, 0],
                "upper_window": [1, 2],
            }
        )
        params.append(
            {"events_df": test_events, "prefix": "test", "dummy_by_window": False}
        )

        test_events = pd.DataFrame(
            {
                "event_name": ["holiday3", "holiday3"],
                "ds": [pd.Timestamp("2020-02-01"), pd.Timestamp("2021-02-10")],
                "lower_window": [-1, 0],
                "upper_window": [1, 2],
            }
        )
        params.append(
            {"events_df": test_events, "prefix": "test", "dummy_by_window": True}
        )
        return params



"""Numpyro inference engines for prophet models.

The classes in this module take a model, the data and perform inference using Numpyro.
"""

from operator import attrgetter
from typing import Callable, Dict, List, Tuple, Union

import jax.numpy as jnp
from jax.random import PRNGKey
from numpyro.diagnostics import summary
from numpyro.infer import MCMC, NUTS, Predictive
from numpyro.infer.initialization import init_to_mean
from numpyro.infer.mcmc import MCMCKernel

from prophetverse.engine.base import BaseInferenceEngine
from prophetverse.engine.utils import assert_mcmc_converged


class MCMCInferenceEngine(BaseInferenceEngine):
    """
    Perform MCMC (Markov Chain Monte Carlo) inference for a given model.

    Parameters
    ----------
    num_samples : int
        The number of MCMC samples to draw.
    num_warmup : int
        The number of warmup samples to discard.
    num_chains : int
        The number of MCMC chains to run in parallel.
    dense_mass : bool
        Whether to use dense mass matrix for NUTS sampler.
    rng_key : Optional
        The random number generator key.
    r_hat: Optional
        The required r_hat for considering the chains to have converged.
    progress_bar : bool
        Whether to show a progress bar during MCMC sampling.

    Attributes
    ----------
    num_samples : int
        The number of MCMC samples to draw.
    num_warmup : int
        The number of warmup samples to discard.
    num_chains : int
        The number of MCMC chains to run in parallel.
    dense_mass : bool
        Whether to use dense mass matrix for NUTS sampler.
    posterior_samples_ : Dict[str, np.ndarray]
        The posterior samples obtained from MCMC.
    samples_predictive_ : Dict[str, np.ndarray]
        The predictive samples obtained from MCMC.
    summary_ : Dict[str, Dict[str, np.ndarray]]
        Summary of statistics for the posterior samples.
    """

    _tags = {
        "inference_method": "mcmc",
    }

    def __init__(
        self,
        num_samples=1000,
        num_warmup=200,
        num_chains=1,
        dense_mass: Union[bool, List[Tuple[str, ...]]] = False,
        rng_key: PRNGKey = None,
        r_hat: Union[float, None] = None,
        progress_bar: bool = True,
    ):
        self.num_samples = num_samples
        self.num_warmup = num_warmup
        self.num_chains = num_chains
        self.dense_mass = dense_mass
        self.r_hat = r_hat
        self.progress_bar = progress_bar

        self.summary_ = None

        super().__init__(rng_key)

    def build_kernel(self, model: Callable) -> MCMCKernel:
        """
        Build the MCMC kernel.

        Parameters
        ----------
        model : Callable
            The model function to perform inference on.

        Returns
        -------
        MCMCKernel
            The MCMC kernel for the model.
        """
        return NUTS(
            model,
            init_strategy=init_to_mean,
            dense_mass=self.dense_mass,
        )

    def _infer(self, **kwargs):
        """
        Run MCMC inference.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the MCMC run method.

        Returns
        -------
        self
            The MCMCInferenceEngine object.
        """

        def get_posterior_samples(
            rng_key, kernel, num_samples, num_warmup, num_chains, progress_bar, **kw
        ) -> Tuple[Dict[str, jnp.ndarray], Dict[str, Dict[str, jnp.ndarray]]]:
            mcmc_ = MCMC(
                kernel,
                num_samples=num_samples,
                num_warmup=num_warmup,
                num_chains=num_chains,
                progress_bar=progress_bar,
            )
            mcmc_.run(rng_key, **kw)

            group_by_chain = True
            samples = mcmc_.get_samples(group_by_chain=group_by_chain)

            # NB: we fetch sample sites to avoid calculating convergence
            # check on deterministic sites, basically same approach
            # as in `mcmc.print_summary`.
            sites = attrgetter(mcmc_._sample_field)(mcmc_._last_state)

            # NB: we keep it simple and only calculate a summary check whenever it
            # satisfies the requirements of split_gelman_rubin. As it's not likely
            # users will use less than four samples.
            if num_samples >= 4:
                filtered_samples = {k: v for k, v in samples.items() if k in sites}
                summary_ = summary(filtered_samples, group_by_chain=group_by_chain)
            else:
                summary_ = {}

            flattened_samples = {
                k: v.reshape((-1,) + v.shape[2:]) for k, v in samples.items()
            }

            return flattened_samples, summary_

        kernel_ = self.build_kernel(self.model_)
        self.posterior_samples_, self.summary_ = get_posterior_samples(
            self._rng_key,
            kernel_,
            num_samples=self.num_samples,
            num_warmup=self.num_warmup,
            num_chains=self.num_chains,
            progress_bar=self.progress_bar,
            **kwargs
        )

        if self.r_hat and self.summary_:
            assert_mcmc_converged(self.summary_, self.r_hat)

        return self

    def _predict(self, **kwargs):
        """
        Generate predictive samples.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the Predictive method.

        Returns
        -------
        Dict[str, np.ndarray]
            The predictive samples.
        """
        num_samples = int(self.num_samples * self.num_chains)

        predictive = Predictive(
            self.model_, self.posterior_samples_, num_samples=num_samples
        )

        self.samples_predictive_ = predictive(self._rng_key, **kwargs)
        return self.samples_predictive_



"""BypassEffect implementation - ignores inputs and returns zeros."""

from typing import Any, Dict, Optional

import jax.numpy as jnp
import pandas as pd

from prophetverse.effects.base import BaseEffect

__all__ = ["IgnoreInput"]


class IgnoreInput(BaseEffect):
    """Effect that ignores all inputs and returns zeros during prediction.

    This effect can be used as a placeholder or to disable specific effects
    without removing them from the model configuration.

    The effect ignores all input data and always returns zeros with the
    appropriate shape for the forecast horizon.

    Parameters
    ----------
    raise_error : bool, optional
        If True, validates that X is empty (has no columns) during fit.
        If False, ignores X completely. Default is False.
    """

    _tags = {
        "requires_X": True,  # Default value, will be overridden in __init__
        "applies_to": "X",
    }

    def __init__(self, raise_error: bool = False):
        """Initialize the BypassEffect.

        Parameters
        ----------
        validate_empty_input : bool, optional
            If True, validates that X is empty (has no columns) during fit.
            If False, ignores X completely. Default is False.
        """
        self.raise_error = raise_error
        super().__init__()

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1.0):
        """Fit the effect. If validation is enabled, check that X is empty.

        Parameters
        ----------
        y : pd.DataFrame
            The target time series data.
        X : pd.DataFrame
            The exogenous variables DataFrame.
        scale : float, optional
            The scale factor, by default 1.0.

        Raises
        ------
        ValueError
            If validate_empty_input is True and X is not empty (has columns).
        """
        if self.raise_error and X is not None and len(X.columns) > 0:
            raise ValueError(
                f"IgnoreInput with raise_error=True requires X to be empty "
                f"(no columns), but X has {len(X.columns)} columns: {list(X.columns)}"
            )

    def _transform(self, X, fh):
        if X is None:
            return None
        return super()._transform(X, fh)

    def _predict(
        self,
        data: Any,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Return zero.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method (ignored).

        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects (ignored).

        Returns
        -------
        jnp.ndarray
            Zero.
        """
        if data is None:
            raise ValueError("Data cannot be None in _predict method.")
        return jnp.zeros((data.shape[0], 1)).astype(jnp.float32)



"""Composition of effects (Effects that wrap other effects)."""

from typing import Any, Dict

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pandas as pd

from prophetverse.utils.frame_to_array import series_to_tensor_or_array

from .base import BaseEffect

__all__ = ["ProxyLikelihood"]


class LinearProxyLikelihood(BaseEffect):
    """
    Use a proxy variable as a likelihood for an effect.

    Assumes there exist an observed variable Z that is correlated with certain
    components we want to predict. This effect adds the likelihood:

    $$
    Z_t \sim \mathcal{N}(\beta f_i, \sigma)
    $$

    Parameters
    ----------
    effect_names : str
        The effect to use in the likelihood.
    reference_df : pd.DataFrame
        A dataframe with the proxy values. Should be in sktime format, and must
        have the same index as the input data.
    coefficient_prior : numpyro.distributions.Distribution, optional
        Prior distribution for the coefficient beta. If None,
          a Normal(1, 0.5) prior is used.
    likelihood_scale : float
        The scale of the likelihood

    """

    _tags = {"requires_X": False, "hierarchical_prophet_compliant": False}

    def __init__(
        self,
        effect_name: str,
        reference_df: pd.DataFrame,
        coefficient_prior=None,
        likelihood_scale: float = 0.05,
        should_rescale: bool = False,
    ):

        self.effect_name = effect_name
        self.reference_df = reference_df
        self.likelihood_scale = likelihood_scale
        self.coefficient_prior = coefficient_prior
        self.should_rescale = should_rescale
        super().__init__()

        assert self.likelihood_scale > 0, "likelihood_scale must be greater than 0"

        self._coefficient_prior = coefficient_prior
        if coefficient_prior is None:
            self._coefficient_prior = dist.Normal(1.0, 0.5)

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1):
        """Initialize the effect.


        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.

        Returns
        -------
        None
        """
        self.scale_ = scale

    def _transform(self, X: pd.DataFrame, fh: pd.Index) -> Dict[str, Any]:
        """Prepare input data to be passed to numpyro model.

        Returns a dictionary with the data for the lift and for the inner effect.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables for the training
            time indexes, if passed during fit, or for the forecasting time indexes, if
            passed during predict.

        fh : pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        Dict[str, Any]
            Dictionary with data for the lift and for the inner effect
        """
        data_dict = {}

        X_ref = self.reference_df.reindex(fh, fill_value=jnp.nan)
        lift_array = series_to_tensor_or_array(X_ref)
        data_dict["observed_reference_value"] = lift_array
        if self.should_rescale:
            data_dict["observed_reference_value"] /= self.scale_
        data_dict["obs_mask"] = ~jnp.isnan(data_dict["observed_reference_value"])
        data_dict["data"] = None
        return data_dict

    def _predict(
        self, data: Dict, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries.
        """
        observed_reference_value = data["observed_reference_value"]
        obs_mask = data["obs_mask"]

        x = predicted_effects[self.effect_name]

        beta = numpyro.sample("beta", self._coefficient_prior)

        with numpyro.handlers.mask(mask=obs_mask):
            numpyro.sample(
                "proxy_likelihood:ignore",
                dist.Normal(x * beta, self.likelihood_scale),
                obs=observed_reference_value,
            )

        return jnp.zeros_like(x)

    @classmethod
    def get_test_params(cls, parameter_set="default"):

        return [
            {
                "effect_name": "trend",
                "reference_df": pd.DataFrame({"y": [1, 2, 3]}),
                "likelihood_scale": 0.1,
            }
        ]



from numpyro.infer import Predictive
from prophetverse.engine.base import BaseInferenceEngine
from numpyro import handlers
from jax import random


class PriorPredictiveInferenceEngine(BaseInferenceEngine):
    """
    Prior‐only inference engine for prior predictive checks.
    Samples parameters from their priors and runs the model.
    """

    _tags = {"inference_method": "prior_predictive"}

    def __init__(
        self, num_samples=1000, rng_key=None, substitute=None, return_sites=None
    ):
        self.num_samples = num_samples
        self.substitute = substitute
        self.return_sites = return_sites
        super().__init__(rng_key)

    def _infer(self, **kwargs):

        _, trace_key, predictive_key = random.split(self._rng_key, 3)

        model = self.model_
        if self.substitute is not None:
            model = handlers.substitute(model, self.substitute)

        trace = handlers.trace(handlers.seed(model, trace_key)).get_trace(**kwargs)
        self.trace_ = trace

        sample_sites = [
            site_name
            for site_name in trace.keys()
            if trace[site_name]["type"] == "sample"
            and not trace[site_name]["is_observed"]
        ]

        prior_predictive = Predictive(
            model,
            num_samples=self.num_samples,
            exclude_deterministic=True,
            return_sites=sample_sites,
        )
        self.posterior_samples_ = prior_predictive(predictive_key, **kwargs)

        if "obs" in self.posterior_samples_:
            # Remove the observed data from the samples
            del self.posterior_samples_["obs"]
        return self

    def _predict(self, **kwargs):
        """
        Draw samples from the prior predictive distribution.
        """
        _, predictive_key = random.split(self._rng_key)
        model = self.model_

        if self.return_sites is None:
            return_sites = None
        elif self.return_sites == "all":
            return_sites = list(self.trace_.keys())
        else:
            return_sites = self.return_sites

        predictive = Predictive(
            model,
            posterior_samples=self.posterior_samples_,
            num_samples=self.num_samples,
            return_sites=return_sites,
        )

        self.samples_predictive_ = predictive(predictive_key, **kwargs)
        return self.samples_predictive_



"""
Piecewise trend models.

This module contains the implementation of piecewise trend models (logistic and linear).
"""

import itertools
from typing import Any, Dict, Tuple, Union, Optional

import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist
import pandas as pd
from sktime.transformations.series.detrend import Detrender

from prophetverse.effects.base import BaseEffect
from prophetverse.utils.frame_to_array import series_to_tensor

from .base import TrendEffectMixin

__all__ = ["PiecewiseLinearTrend", "PiecewiseLogisticTrend"]


class PiecewiseLinearTrend(TrendEffectMixin, BaseEffect):
    """Piecewise Linear Trend model.

    This model assumes that the trend is piecewise linear, with changepoints
    at regular intervals. The number of changepoints is determined by the
    `changepoint_interval` and `changepoint_range` parameters. The
    `changepoint_interval` parameter specifies the interval between changepoints,
    while the `changepoint_range` parameter specifies the range of the changepoints.

    This implementation is based on the `Prophet`_ library. The initial values (global
    rate and global offset) are suggested using the maximum and minimum values of the
    time series data.


    Parameters
    ----------
    changepoint_interval : int
        The interval between changepoints.
    changepoint_range : int
        The range of the changepoints.
    changepoint_prior_scale : dist.Distribution
        The prior scale for the changepoints.
    offset_prior_scale : float, optional
        The prior scale for the offset. Default is 0.1.
    remove_seasonality_before_suggesting_initial_vals : bool, optional
        If True, remove seasonality before suggesting initial values, using sktime's
        detrender. Default is True.
    global_rate_prior_loc : float, optional
        The prior location for the global rate. Default is suggested
        empirically from data.
    offset_prior_loc : float, optional
        The prior location for the offset. Default is suggested
        empirically from data.


    """

    _tags = {"capability:panel": False}

    def __init__(
        self,
        changepoint_interval: int = 25,
        changepoint_range: int = 0.8,
        changepoint_prior_scale: dist.Distribution = 0.001,
        offset_prior_scale=0.1,
        remove_seasonality_before_suggesting_initial_vals: bool = True,
        global_rate_prior_loc: Optional[float] = None,
        offset_prior_loc: Optional[float] = None,
    ):
        self.changepoint_interval = changepoint_interval
        self.changepoint_range = changepoint_range
        self.changepoint_prior_scale = changepoint_prior_scale
        self.offset_prior_scale = offset_prior_scale
        self.remove_seasonality_before_suggesting_initial_vals = (
            remove_seasonality_before_suggesting_initial_vals
        )
        self.global_rate_prior_loc = global_rate_prior_loc
        self.offset_prior_loc = offset_prior_loc
        super().__init__()

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1):
        """Initialize the effect.

        Set the prior location for the trend.

        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.
        """
        super()._fit(X=X, y=y, scale=scale)

        t_scaled = self._index_to_scaled_timearray(
            y.index.get_level_values(-1).unique()
        )
        self._setup_changepoints(t_scaled)
        self._setup_changepoint_prior_vectors(y)
        self._index_names = y.index.names

    def _transform(self, X: pd.DataFrame, fh: pd.Index) -> dict:
        """
        Prepare the input data for the piecewise trend model.

        Parameters
        ----------
        X: pd.DataFrame
            The exogenous variables DataFrame.
        fh: pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        jnp.ndarray
            An array containing the prepared input data.
        """
        return self.get_changepoint_matrix(fh)

    def _predict(
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Compute the trend based on the given changepoint matrix.

        Parameters
        ----------
        data: jnp.ndarray
            The changepoint matrix.
        predicted_effects: Dict[str, jnp.ndarray]
            Dictionary of previously computed effects. For the trend, it is an empty
            dict.

        Returns
        -------
        jnp.ndarray
            The computed trend.
        """
        # alias for clarity
        changepoint_matrix = data

        offset = numpyro.sample(
            "offset",
            dist.Normal(self._offset_prior_loc, self._offset_prior_scale),
        )

        changepoint_coefficients = numpyro.sample(
            "changepoint_coefficients",
            dist.Laplace(self._changepoint_prior_loc, self._changepoint_prior_scale),
        )

        trend = (changepoint_matrix) @ changepoint_coefficients + offset

        return trend.reshape((-1, 1))

    def get_changepoint_matrix(self, idx: pd.PeriodIndex) -> jnp.ndarray:
        """
        Return the changepoint matrix for the given index.

        Parameters
        ----------
        idx: pd.PeriodIndex
            The index for which to compute the changepoint matrix.

        Returns
        -------
            jnp.ndarray: The changepoint matrix.
        """
        t_scaled = self._index_to_scaled_timearray(idx)
        changepoint_matrix = _get_changepoint_matrix(t_scaled, self._changepoint_ts)

        return changepoint_matrix

    @property
    def n_changepoint_per_series(self):
        """Get the number of changepoints per series.

        Returns
        -------
        int
            Number of changepoints per series.
        """
        return [len(cp) for cp in self._changepoint_ts]

    @property
    def n_changepoints(self):
        """Get the total number of changepoints.

        Returns
        -------
        int
            Total number of changepoints.
        """
        return len(self._changepoint_ts)

    def _setup_changepoints(self, t_scaled) -> None:
        """
        Set changepoint variables.

        This function has the collateral effect of setting the following attributes:
        - self._changepoint_ts.

        Parameters
        ----------
        t_scaled: jnp.ndarray
            Transformed time index.

        Returns
        -------
            None
        """

        self._changepoint_ts = _get_changepoint_timeindexes(
            t_scaled,
            changepoint_interval=self.changepoint_interval,
            changepoint_range=self.changepoint_range,
        )

        if len(self._changepoint_ts) == 0:
            raise ValueError(
                "No changepoints were generated. Try increasing the changing"
                + f" the changepoint_range. There are {len(t_scaled)} timepoints "
                + f" in the series, changepoint_range is {self.changepoint_range} and "
                + f"changepoint_interval is {self.changepoint_interval}."
            )

    def _setup_changepoint_prior_vectors(self, y: pd.DataFrame) -> None:
        """
        Set up the changepoint prior vectors for the model.

        Parameters
        ----------
        y: pd.DataFrame
            The input DataFrame containing the time series data.

        Returns
        -------
            None
        """
        if self.remove_seasonality_before_suggesting_initial_vals:
            detrender = Detrender()
            y = y - detrender.fit_transform(y)

        self._global_rate, self._offset_prior_loc = (
            self._suggest_global_trend_and_offset(y)
        )
        self._changepoint_prior_loc, self._changepoint_prior_scale = (
            self._get_changepoint_prior_vectors(global_rate=self._global_rate)
        )
        self._offset_prior_scale = self.offset_prior_scale

    def _get_changepoint_prior_vectors(
        self,
        global_rate: jnp.array,
    ) -> Tuple[jnp.array, jnp.array]:
        """
        Return the prior vectors for the changepoint coefficients.

        Parameters
        ----------
        global_rates: jnp.array
            The global rates for each series.

        Returns
        -------
        Tuple[jnp.array, jnp.array]
            A tuple containing the changepoint prior location vector and the
            changepoint prior scale vector.
        """

        def zeros_with_first_value(size, first_value):
            x = jnp.zeros(size)
            x = x.at[0].set(first_value)
            return x

        changepoint_prior_scale_vector = (
            np.ones(self.n_changepoints) * self.changepoint_prior_scale
        )

        changepoint_prior_loc_vector = zeros_with_first_value(
            self.n_changepoints, global_rate
        )

        return jnp.array(changepoint_prior_loc_vector), jnp.array(
            changepoint_prior_scale_vector
        )

    def _suggest_global_trend_and_offset(
        self, y: pd.DataFrame
    ) -> Tuple[jnp.array, jnp.array]:
        """
        Suggest the global trend and offset for the given time series data.

        Parameters
        ----------
        y: pd.DataFrame
            The time series data.

        Returns
        -------
        Tuple[jnp.array, jnp.array]
            A tuple containing the global trend and offset.
        """
        t = self._index_to_scaled_timearray(y.index.get_level_values(-1).unique())

        y_array = series_to_tensor(y)

        global_rate = _enforce_array_if_zero_dim(
            (y_array[:, -1].squeeze() - y_array[:, 0].squeeze())
            / (t[0].squeeze() - t[-1].squeeze())
        )
        offset_loc = y_array[:, 0].squeeze() - global_rate * t[0].squeeze()

        global_rate = global_rate.item()
        offset_loc = offset_loc.item()

        if self.offset_prior_loc is not None:
            offset_loc = self.offset_prior_loc

        if self.global_rate_prior_loc is not None:
            global_rate = self.global_rate_prior_loc

        return global_rate, offset_loc


class PiecewiseLogisticTrend(PiecewiseLinearTrend):
    """
    Piecewise logistic trend model.

    This logistic trend differs from the original Prophet logistic trend in that it
    considers a capacity prior distribution. The capacity prior distribution is used
    to estimate the maximum value that the time series trend can reach.

    It uses internally the piecewise linear trend model, and then applies a logistic
    function to the output of the linear trend model.


    The initial values (global rate and global offset) are suggested using the maximum
    and minimum values of the time series data.


    Parameters
    ----------
    changepoint_interval : int
        The interval between changepoints.
    changepoint_range : int
        The range of the changepoints.
    changepoint_prior_scale : dist.Distribution
        The prior scale for the changepoints.
    offset_prior_scale : float, optional
        The prior scale for the offset. Default is 0.1.
    remove_seasonality_before_suggesting_initial_vals : bool, optional
        If True, remove seasonality before suggesting initial values, using sktime's
        detrender. Default is True.
    capacity_prior : dist.Distribution, optional
        The prior distribution for the capacity. Default is a HalfNormal distribution
        with loc=1.05 and scale=1.
    """

    def __init__(
        self,
        changepoint_interval: int = 25,
        changepoint_range: int = 0.8,
        changepoint_prior_scale: float = 0.001,
        offset_prior_scale=10,
        capacity_prior: dist.Distribution = None,
        remove_seasonality_before_suggesting_initial_vals: bool = True,
        global_rate_prior_loc: Optional[float] = None,
        offset_prior_loc: Optional[float] = None,
    ):

        self.capacity_prior = capacity_prior

        super().__init__(
            changepoint_interval,
            changepoint_range,
            changepoint_prior_scale,
            offset_prior_scale=offset_prior_scale,
            remove_seasonality_before_suggesting_initial_vals=remove_seasonality_before_suggesting_initial_vals,
            global_rate_prior_loc=global_rate_prior_loc,
            offset_prior_loc=offset_prior_loc,
        )

        if capacity_prior is None:
            capacity_prior = dist.TransformedDistribution(
                dist.HalfNormal(
                    0.2,
                ),
                dist.transforms.AffineTransform(loc=1.1, scale=1),
            )
        self._capacity_prior = capacity_prior

    def _suggest_global_trend_and_offset(
        self, y: pd.DataFrame
    ) -> Tuple[jnp.array, jnp.array]:
        """
        Suggest the global trend and offset for the given time series data.

        This implementation considers the max and min values of the timeseries
        to anallytically estimate the global rate and offset.

        Parameters
        ----------
        y: pd.DataFrame
            The input time series data.

        Returns
        -------
        Tuple[jnp.array, jnp.array]
            A tuple containing the suggested global rates
            and offset.
        """
        t_arrays = self._index_to_scaled_timearray(
            y.index.get_level_values(-1).unique()
        )
        y_arrays = series_to_tensor(y)

        if hasattr(self._capacity_prior, "loc"):
            capacity_prior_loc = self._capacity_prior.loc
        else:
            capacity_prior_loc = y_arrays.max() * 1.05

        global_rates, offset = _suggest_logistic_rate_and_offset(
            t=t_arrays.squeeze(),
            y=y_arrays.squeeze(),
            capacities=capacity_prior_loc,
        )

        return global_rates.item(), offset.item()

    def _predict(  # type: ignore[override]
        self, data: Any, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        """
        Compute the trend for the given changepoint matrix.

        Parameters
        ----------
        changepoint_matrix: jnp.ndarray
            The changepoint matrix.

        Returns
        -------
        jnp.ndarray
            The computed trend.
        """
        with numpyro.plate("series", self.n_series, dim=-3):
            capacity = numpyro.sample("capacity", self._capacity_prior)

        trend = super()._predict(data=data, predicted_effects=predicted_effects)

        if self.n_series == 1:
            capacity = capacity.squeeze()

        trend = capacity / (1 + jnp.exp(-trend))
        return trend


def _to_list_if_scalar(x, size=1):
    """
    Convert a scalar value to a list of the same value repeated `size` times.

    Parameters
    ----------
    x: scalar or array-like
        The input value to be converted.
    size: int, optional
        The number of times to repeat the value in the list. Default is 1.

    Returns
    -------
    list
        A list containing the input value repeated `size` times if `x` is
        a scalar, otherwise returns `x` unchanged.
    """
    if np.isscalar(x):
        return [x] * size
    return x


def _enforce_array_if_zero_dim(x):
    """
    Reshapes the input array `x` to have at least one dimension if it has zero dim.

    Parameters
    ----------
    x: array-like
        The input array.

    Returns
    -------
    array-like
        The reshaped array.
    """
    if x.ndim == 0:
        return x.reshape(1)
    return x


def _suggest_logistic_rate_and_offset(
    t: np.ndarray, y: np.ndarray, capacities: Union[float, np.ndarray]
):
    """
    Suggest the logistic rate and offset based on the given time series data.

    Parameters
    ----------
    t: ndarray
        The time values of the time series data.
    y: ndarray
        The observed values of the time series data.
    capacities: float or ndarray
        The capacity or capacities of the time series data.

    Returns
    -------
    m: jnp.ndarray
        The suggested offset.
    k: jnp.ndarray
        The suggested logistic rate.
    """
    if y.ndim == 1:
        y = y.reshape(1, -1)
    elif y.ndim == 3:
        # Shape here would be (n_series, n_samples, 1)
        y = y.squeeze()
    if t.ndim == 1:
        t = t.reshape(1, -1)
    elif t.ndim == 3:
        # Shape here would be (n_series, n_samples, 1)
        t = t.squeeze()

    i0, i1 = t.argmin(axis=1), t.argmax(axis=1)
    t0, t1 = t[:, i0].flatten(), t[:, i1].flatten()
    T = t0 - t1
    y0, y1 = y[:, i0].flatten(), y[:, i1].flatten()

    if np.any(y0 == 0):
        y0 = 1e-8

    r0 = capacities / y0
    r1 = capacities / y1

    L0 = np.log(r0 - 1)
    L1 = np.log(r1 - 1)

    if np.any(np.isnan(L0)) or np.any(np.isnan(L1)):
        error_str = (
            "L0 is {}, L1 is {}. At least one of them has NaN,"
            + "the input parameters were: t: {}, y: {}, capacities: {}"
        )
        raise ValueError(error_str.format(L0, L1, t, y, capacities))

    k = (L1 - L0) / T
    m = -(L1 + k * t1)

    return k, m


def _get_changepoint_matrix(t: jnp.ndarray, changepoint_t: jnp.array) -> jnp.ndarray:
    """
    Generate a changepoint matrix based on the time indexes and changepoint indexes.

    Parameters
    ----------
    t: jnp.ndarray
        array with timepoints of shape (n, 1) preferably
    changepoint_t: jnp.array
        array with changepoint timepoints of shape (n_changepoints,)

    Returns
    -------
    jnp.ndarray
        changepoint matrix of shape (n, n_changepoints)
    """
    expanded_ts = jnp.tile(t.reshape((-1, 1)), (1, len(changepoint_t)))
    A = (expanded_ts >= changepoint_t.reshape((1, -1))).astype(int) * expanded_ts
    cutoff_ts = changepoint_t.reshape((1, -1))
    A = jnp.clip(A - cutoff_ts, 0, None)
    return A


def _get_changepoint_timeindexes(
    t: jnp.ndarray, changepoint_interval: int, changepoint_range: float = 0.90
) -> jnp.array:
    """
    Return an array of time indexes for changepoints based on the given parameters.

    Parameters
    ----------
    t: jnp.ndarray
        The array of time values.
    changepoint_interval: int
        The interval between changepoints.
    changepoint_range: float, optional
        The range of changepoints. Defaults to 0.90. If greater than 1, it is
        interpreted as then number of timepoints.If less than zero, it is interpreted as
        number of timepoints from the end of the time series.

    Returns
    -------
    jnp.array
        An array of time indexes for changepoints.
    """
    if changepoint_range < 1 and changepoint_range > 0:
        max_t = t.max() * changepoint_range
    elif changepoint_range >= 1:
        max_t = changepoint_range
    else:
        max_t = t.max() + changepoint_range

    changepoint_t = jnp.arange(0, max_t, changepoint_interval)
    return changepoint_t



from .sktime import Prophetverse
from .effects import (
    # Trend effects
    FlatTrend,
    PiecewiseLinearTrend,
    PiecewiseLogisticTrend,
    # Target likelihoods
    NormalTargetLikelihood,
    MultivariateNormal,
    GammaTargetLikelihood,
    NegativeBinomialTargetLikelihood,
    BetaTargetLikelihood,
    # Exogenous effects
    MultiplyEffects,
    MichaelisMentenEffect,
    HillEffect,
    LinearEffect,
    LinearFourierSeasonality,
    LiftExperimentLikelihood,
    ExactLikelihood,
    GeometricAdstockEffect,
    WeibullAdstockEffect,
    ChainedEffects,
    Forward,
    IgnoreInput,
    Identity,
    Constant,
    CoupledExactLikelihood,
)
from .engine import (
    MAPInferenceEngine,
    MCMCInferenceEngine,
    PriorPredictiveInferenceEngine,
    VIInferenceEngine,
)

from .engine.optimizer import (
    LBFGSSolver,
    CosineScheduleAdamOptimizer,
    AdamOptimizer,
)


from .budget_optimization import BudgetOptimizer
from .budget_optimization.constraints import (
    TotalBudgetConstraint,
    SharedBudgetConstraint,
    MinimumTargetResponse,
)
from .budget_optimization.objectives import MaximizeKPI, MaximizeROI, MinimizeBudget

from .budget_optimization.parametrization_transformations import (
    InvestmentPerChannelAndSeries,
    InvestmentPerChannelTransform,
    TotalInvestmentTransform,
    InvestmentPerSeries,
    IdentityTransform,
)

__all__ = [
    "Prophetverse",
    # Effects
    "FlatTrend",
    "PiecewiseLinearTrend",
    "PiecewiseLogisticTrend",
    "NormalTargetLikelihood",
    "MultivariateNormal",
    "GammaTargetLikelihood",
    "NegativeBinomialTargetLikelihood",
    "MultiplyEffects",
    "MichaelisMentenEffect",
    "HillEffect",
    "LinearEffect",
    "LinearFourierSeasonality",
    "LiftExperimentLikelihood",
    "ExactLikelihood",
    "GeometricAdstockEffect",
    "WeibullAdstockEffect",
    "ChainedEffects",
    "BetaTargetLikelihood",
    "IgnoreInput",
    "Identity",
    "Constant",
    "CoupledExactLikelihood",
    # Engine
    "MAPInferenceEngine",
    "MCMCInferenceEngine",
    "PriorPredictiveInferenceEngine",
    "VIInferenceEngine",
    # Optimizers
    "LBFGSSolver",
    "CosineScheduleAdamOptimizer",
    "AdamOptimizer",
    # Budget Optimization
    "BudgetOptimizer",
    "TotalBudgetConstraint",
    "SharedBudgetConstraint",
    "MinimumTargetResponse",
    "MaximizeKPI",
    "MaximizeROI",
    "MinimizeBudget",
    "InvestmentPerChannelAndSeries",
    "InvestmentPerChannelTransform",
    "TotalInvestmentTransform",
    "InvestmentPerSeries",
    "IdentityTransform",
    "Forward",
]



import re
from typing import Optional

import numpy as np
import pandas as pd
from sktime.transformations.base import BaseTransformer


class ExpandColumnPerLevel(BaseTransformer):
    """
    Expand columns based on a regular expression pattern per level of a multi-level idx.

    If there are M columns matching the pattern and N distinct multi-level indexes, then
    the transformer will create M * N new columns, setting the value to zero to each of
    M * (N-1) columns for a given index j if the current index i /= j.

    Parameters
    ----------
    columns_regex : list
        A list of regular expression patterns to match the columns to be expanded.

    Attributes
    ----------
    matched_columns_ : set
        A set of columns that match the regular expression patterns.
    new_columns_ : dict
        A dictionary mapping the new column names to the original column names.

    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input data.
    transform(X, y=None)
        Transform the input data by expanding the columns.
    """

    def __init__(self, columns_regex: list[str]):
        self.columns_regex = columns_regex
        super().__init__()

    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None):
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : pd.DataFrame
            The input data.
        y : pd.Series or None
            The target variable. Not used in this transformer.

        Returns
        -------
        self : ExpandColumnPerLevel
            The fitted transformer object.
        """
        regex_patterns = [re.compile(str(pattern)) for pattern in self.columns_regex]
        self.matched_columns_ = {
            col
            for col in X.columns
            for pattern in regex_patterns
            if pattern.match(str(col))
        }

        self.new_columns_ = {}
        # Ensure identifiers are tuples
        if X.index.nlevels == 1:
            self.skip_transform_ = True
            self.series_identifiers_ = None
            self.new_columns_ = {col: col for col in self.matched_columns_}
            return self

        self.skip_transform_ = False
        self.series_identifiers_ = [
            idx if isinstance(idx, tuple) else (idx,)
            for idx in X.index.droplevel(-1).unique()
        ]
        for identifier in self.series_identifiers_:
            for col in self.matched_columns_:
                self.new_columns_[self.get_col_name(col, identifier)] = col
        return self

    @classmethod
    def get_col_name(cls, column: str, identifier: tuple) -> str:
        """
        Generate the new column name based on the original name and the identifier.

        Parameters
        ----------
        column : str
            The original column name.
        identifier : tuple
            The identifier for the column.

        Returns
        -------
        new_col_name : str
            The new column name.
        """
        return f"{column}_dup_{'-'.join(identifier)}"

    def transform(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None):
        """
        Transform the input data by expanding the columns.

        Parameters
        ----------
        X : pd.DataFrame
            The input data.
        y : pd.Series or None
            The target variable. Not used in this transformer.

        Returns
        -------
        X_transformed : pd.DataFrame
            The transformed data with expanded columns.
        """
        if self.skip_transform_:
            return X

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("Input must have a multi-level index")
        X_original = X.copy()
        X = X.copy()

        matched_columns = self.matched_columns_

        # Ensure identifiers are tuples
        series_identifiers = [
            idx if isinstance(idx, tuple) else (idx,)
            for idx in X.index.droplevel(-1).unique()
        ]

        def get_col_name(column, identifier):
            return f"{column}_dup_{'-'.join(identifier)}"

        for match in matched_columns:
            for identifier in series_identifiers:
                X[get_col_name(match, identifier)] = np.float64(0)

        self.new_columns_ = {}
        for identifier in series_identifiers:

            for column in matched_columns:
                self.new_columns_[get_col_name(column, identifier)] = column
                X.loc[identifier, get_col_name(column, identifier)] = X_original.loc[
                    identifier, column
                ].values.flatten()
                X[get_col_name(column, identifier)] = X[
                    get_col_name(column, identifier)
                ].fillna(0)

        return X.drop(columns=matched_columns)

    @classmethod
    def get_test_params(cls, parameter_set="default"):

        return [
            {"columns_regex": ["*"]},
        ]



"""Custom numpyro distributions.

This module contains custom distributions that can be used as likelihoods or priors
for models.
"""

from .reparametrization import GammaReparametrized, BetaReparametrized
from .hurdle_distribution import HurdleDistribution
from .truncated_discrete import TruncatedDiscrete

__all__ = [
    "GammaReparametrized",
    "BetaReparametrized",
    "HurdleDistribution",
    "TruncatedDiscrete",
]



"""Datasets for ProphetVerse."""

from .loaders import (
    load_forecastingdata,
    load_pedestrian_count,
    load_peyton_manning,
    load_tensorflow_github_stars,
    load_tourism,
)
from .synthetic import load_composite_effect_example, load_synthetic_squared_exogenous

__all__ = [
    "load_forecastingdata",
    "load_pedestrian_count",
    "load_peyton_manning",
    "load_tensorflow_github_stars",
    "load_tourism",
    "load_synthetic_squared_exogenous",
    "load_composite_effect_example",
]



"""Definition of Hill Effect class."""

from typing import Dict, Optional, Any

import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
from numpyro.distributions import Distribution

from prophetverse.effects.base import (
    EFFECT_APPLICATION_TYPE,
    BaseAdditiveOrMultiplicativeEffect,
)
from prophetverse.utils.algebric_operations import _exponent_safe

__all__ = ["HillEffect"]


class HillEffect(BaseAdditiveOrMultiplicativeEffect):
    """Represents a Hill effect in a time series model.

    Parameters
    ----------
    half_max_prior : Distribution, optional
        Prior distribution for the half-maximum parameter
    slope_prior : Distribution, optional
        Prior distribution for the slope parameter
    max_effect_prior : Distribution, optional
        Prior distribution for the maximum effect parameter
    effect_mode : effects_application, optional
        Mode of the effect (either "additive" or "multiplicative")
    """

    def __init__(
        self,
        effect_mode: EFFECT_APPLICATION_TYPE = "multiplicative",
        half_max_prior: Optional[Distribution] = None,
        slope_prior: Optional[Distribution] = None,
        max_effect_prior: Optional[Distribution] = None,
        offset_slope: Optional[float] = 0.0,
        input_scale: Optional[float] = 1.0,
        base_effect_name="trend",
    ):
        self.half_max_prior = half_max_prior
        self.slope_prior = slope_prior
        self.max_effect_prior = max_effect_prior

        self._half_max_prior = (
            self.half_max_prior if half_max_prior is not None else dist.Gamma(1, 1)
        )
        self._slope_prior = (
            self.slope_prior if slope_prior is not None else dist.HalfNormal(10)
        )
        self._max_effect_prior = (
            self.max_effect_prior if max_effect_prior is not None else dist.Gamma(1, 1)
        )
        self.offset_slope = offset_slope
        self.input_scale = input_scale

        super().__init__(effect_mode=effect_mode, base_effect_name=base_effect_name)

    def _predict(
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries.
        """
        half_max = numpyro.sample("half_max", self._half_max_prior) * self.input_scale
        slope = numpyro.sample("slope", self._slope_prior) + self.offset_slope
        max_effect = numpyro.sample("max_effect", self._max_effect_prior)

        data = jnp.clip(data, 1e-9, None)
        x = _exponent_safe(data / half_max, -slope)
        effect = max_effect / (1 + x)
        return effect



"""Utilities for converting a pandas DataFrame to JAX/Numpy arrays."""

from typing import Tuple

import numpy as np
import pandas as pd
from jax import numpy as jnp

from .multiindex import iterate_all_series

NANOSECONDS_TO_SECONDS = 1000 * 1000 * 1000

__all__ = [
    "convert_index_to_days_since_epoch",
    "series_to_tensor",
    "extract_timetensor_from_dataframe",
    "convert_dataframe_to_tensors",
]


def convert_index_to_days_since_epoch(idx: pd.Index) -> np.array:
    """
    Convert a pandas Index to days since epoch.

    Parameters
    ----------
    idx : pd.Index
        The pandas Index.

    Returns
    -------
    np.ndarray
        The converted array of days since epoch.
    """
    t = idx

    if not (isinstance(t, pd.PeriodIndex) or isinstance(t, pd.DatetimeIndex)):
        return t.values

    if isinstance(t, pd.PeriodIndex):
        t = t.to_timestamp()

    return t.to_numpy(dtype=np.int64) // NANOSECONDS_TO_SECONDS / (3600 * 24.0)


def series_to_tensor(y: pd.DataFrame) -> jnp.ndarray:
    """
    Convert all series of a hierarchical time series to a JAX tensor.

    Parameters
    ----------
    y : pd.DataFrame
        The hierarchical time series.

    Returns
    -------
    jnp.ndarray
        The JAX tensor representing all series.
    """
    names = []
    array = []
    series_len = None

    if y.index.nlevels == 1:
        return jnp.array(y.values).reshape((1, -1, len(y.columns)))

    for idx, series in iterate_all_series(y):
        if series_len is None:
            series_len = len(series)
        if len(series) != series_len:
            raise ValueError(
                f"Series {idx} has length {len(series)}, but expected {series_len}"
            )

        names.append(idx)
        array.append(series.values.reshape((-1, len(y.columns))))
    return jnp.array(array)


def series_to_tensor_or_array(y: pd.DataFrame) -> jnp.ndarray:
    """
    Convert hierarchical (univariate) to three (two) dimensional JAX tensor.

    Parameters
    ----------
    y : pd.DataFrame
        The hierarchical time series.

    Returns
    -------
    jnp.ndarray
        The JAX tensor or the array representing all series.
    """
    if y.index.nlevels == 1:
        return jnp.array(y.values)
    return series_to_tensor(y)


def extract_timetensor_from_dataframe(df: pd.DataFrame) -> jnp.array:
    """
    Extract the time array from a pandas DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame.

    Returns
    -------
    jnp.ndarray
        The JAX tensor representing the time array.
    """
    return series_to_tensor(
        pd.DataFrame(
            index=df.index,
            data={
                "t": convert_index_to_days_since_epoch(df.index.get_level_values(-1))
            },
        )
    )


def convert_dataframe_to_tensors(df: pd.DataFrame) -> Tuple[jnp.array, jnp.array]:
    """
    Convert a pandas DataFrame to JAX tensors.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame.

    Returns
    -------
    tuple
        A tuple containing the time arrays and the DataFrame arrays as JAX tensors.
    """
    t_arrays = extract_timetensor_from_dataframe(df)

    df_as_arrays = series_to_tensor(df)

    return t_arrays, df_as_arrays



"""Chained effects utilities.

Contains a shared base class for chaining multiple effects and two concrete
implementations: `Sum` (additive chaining) and `MultiplyEffects` (multiplicative
chaining). This consolidates duplicated logic into a single file.
"""

from typing import Any, Dict, List, Tuple

import jax.numpy as jnp
from skbase.base import BaseMetaEstimatorMixin
import numpyro
from prophetverse.effects.base import BaseEffect

__all__ = ["Sum", "MultiplyEffects"]


class _BaseChainOperation(BaseMetaEstimatorMixin, BaseEffect):
    """Base class with shared logic for chaining multiple effects.

    Subclasses must implement `_predict` to define how per-effect
    predictions are aggregated (e.g., sum or product).
    """

    _tags = {
        # Can handle panel data?
        "capability:panel": True,
        # Can be used with hierarchical Prophet?
        "capability:panel": True,
        # Can handle multiple input feature columns?
        "capability:multivariate_input": True,
        # If no columns are found, should
        # _predict be skipped?
        "requires_X": False,
        # Should only the indexes related to the forecasting horizon be passed to
        "filter_indexes_with_forecating_horizon_at_transform": True,
        "named_object_parameters": "named_effects",
    }

    def __init__(self, effects: List[Tuple[str, BaseEffect]]):
        self.effects = effects
        super().__init__()

        # Normalize to list of (name, effect)
        self.named_effects = []
        for i, val in enumerate(self.effects):
            if isinstance(val, tuple):
                self.named_effects.append(val)
            elif isinstance(val, BaseEffect):
                self.named_effects.append((str(i), val))
            else:  # pragma: no cover
                raise ValueError(
                    f"Invalid type {type(val)} for step {i}. Must be a tuple or BaseEffect."
                )

        requires_X = any(
            effect.get_tag("requires_X", False) for _, effect in self.named_effects
        )
        self.set_tags(requires_X=requires_X)

    def _fit(self, y: Any, X: Any, scale: float = 1.0):
        """Fit all chained effects sequentially."""
        self.named_effects_ = []
        for name, effect in self.named_effects:
            effect = effect.clone()
            effect.fit(y, X, scale)
            self.named_effects_.append((name, effect))

    def _transform(self, X: Any, fh: Any) -> Any:
        """Transform input data sequentially through all chained effects."""
        all_data = {
            name: effect.transform(X=X, fh=fh) for name, effect in self.named_effects_
        }
        return all_data

    def _update_data(self, data, arr):
        for name, effect in self.named_effects_:
            data[name] = effect._update_data(data[name], arr)
        return data

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        from prophetverse.effects.linear import LinearEffect

        return [
            {
                "effects": [
                    ("linear1", LinearEffect()),
                    ("linear2", LinearEffect()),
                ]
            },
            {"effects": [LinearEffect()]},
        ]

    def _predict(
        self, *args, **kwargs
    ) -> jnp.ndarray:  # pragma: no cover - implemented by subclasses
        raise NotImplementedError()


class MultiplyEffects(_BaseChainOperation):
    """Chains multiple effects multiplicatively."""

    def _predict(
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        output = 1
        for name, effect in self.named_effects_:
            with numpyro.handlers.scope(prefix=name):
                output *= effect.predict(data[name], predicted_effects)

        return output


class SumEffects(_BaseChainOperation):
    """Chains multiple effects additively (sums their contributions)."""

    def _predict(
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        output = 0
        for name, effect in self.named_effects_:
            with numpyro.handlers.scope(prefix=name):
                output = output + effect.predict(data[name], predicted_effects)

        return output



"""Definition of Michaelis-Menten Effect class."""

from typing import Dict, Optional

import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
from numpyro.distributions import Distribution

from prophetverse.effects.base import (
    EFFECT_APPLICATION_TYPE,
    BaseAdditiveOrMultiplicativeEffect,
)

__all__ = ["MichaelisMentenEffect"]


class MichaelisMentenEffect(BaseAdditiveOrMultiplicativeEffect):
    """Represents a Michaelis-Menten effect in a time series model.

    The Michaelis-Menten equation is commonly used in biochemistry to describe
    enzyme kinetics, but it's also useful for modeling saturation effects in
    time series analysis. The effect follows the equation:

    effect = (max_effect * data) / (half_saturation + data)

    Where:
    - max_effect is the maximum effect value (Vmax in biochemistry)
    - half_saturation is the value at which effect = max_effect/2 (Km in biochemistry)
    - data is the input variable (substrate concentration in biochemistry)

    Parameters
    ----------
    max_effect_prior : Distribution, optional
        Prior distribution for the maximum effect parameter
    half_saturation_prior : Distribution, optional
        Prior distribution for the half-saturation parameter
    effect_mode : effects_application, optional
        Mode of the effect (either "additive" or "multiplicative")
    """

    def __init__(
        self,
        effect_mode: EFFECT_APPLICATION_TYPE = "multiplicative",
        max_effect_prior: Optional[Distribution] = None,
        half_saturation_prior: Optional[Distribution] = None,
        base_effect_name="trend",
    ):
        self.max_effect_prior = max_effect_prior
        self.half_saturation_prior = half_saturation_prior

        self._max_effect_prior = (
            self.max_effect_prior if max_effect_prior is not None else dist.Gamma(1, 1)
        )
        self._half_saturation_prior = (
            self.half_saturation_prior if half_saturation_prior is not None else dist.Gamma(1, 1)
        )

        super().__init__(effect_mode=effect_mode, base_effect_name=base_effect_name)

    def _predict(
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries.
        """
        max_effect = numpyro.sample("max_effect", self._max_effect_prior)
        half_saturation = numpyro.sample("half_saturation", self._half_saturation_prior)

        # Clip data to avoid numerical issues with very small values
        data = jnp.clip(data, 1e-9, None)
        
        # Apply Michaelis-Menten equation: effect = (max_effect * data) / (half_saturation + data)
        effect = (max_effect * data) / (half_saturation + data)
        
        return effect


"""Composition of effects (Effects that wrap other effects)."""

from typing import Any, Dict

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pandas as pd

from prophetverse.utils.frame_to_array import series_to_tensor_or_array

from .base import BaseEffect

__all__ = ["ExactLikelihood"]


class ExactLikelihood(BaseEffect):
    """Wrap an effect and applies a normal likelihood to its output.

    This class uses an input as a reference for the effect, and applies a normal
    likelihood to the output of the effect.

    Parameters
    ----------
    effect_name : str
        The effect to use in the likelihood.
    reference_df : pd.DataFrame
        A dataframe with the reference values. Should be in sktime format, and must
        have the same index as the input data.
    prior_scale : float
        The scale of the prior distribution for the likelihood.
    """

    _tags = {"requires_X": False, "hierarchical_prophet_compliant": False}

    def __init__(
        self,
        effect_name: str,
        reference_df: pd.DataFrame,
        prior_scale: float,
    ):

        self.effect_name = effect_name
        self.reference_df = reference_df
        self.prior_scale = prior_scale

        assert self.prior_scale > 0, "prior_scale must be greater than 0"

        super().__init__()

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1):
        """Initialize the effect.

        This method is called during `fit()` of the forecasting model.
        It receives the Exogenous variables DataFrame and should be used to initialize
        any necessary parameters or data structures, such as detecting the columns that
        match the regex pattern.

        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.

        scale : float, optional
            The scale of the timeseries. For multivariate timeseries, this is
            a dataframe. For univariate, it is a simple float.

        Returns
        -------
        None
        """
        self.timeseries_scale = scale

    def _transform(self, X: pd.DataFrame, fh: pd.Index) -> Dict[str, Any]:
        """Prepare input data to be passed to numpyro model.

        Returns a dictionary with the data for the lift and for the inner effect.

        Parameters
        ----------
        X : pd.DataFrame
            The input DataFrame containing the exogenous variables for the training
            time indexes, if passed during fit, or for the forecasting time indexes, if
            passed during predict.

        fh : pd.Index
            The forecasting horizon as a pandas Index.

        Returns
        -------
        Dict[str, Any]
            Dictionary with data for the lift and for the inner effect
        """
        data_dict = {}

        X_lift = self.reference_df.reindex(fh, fill_value=jnp.nan)
        lift_array = series_to_tensor_or_array(X_lift)
        data_dict["observed_reference_value"] = lift_array / self.timeseries_scale
        data_dict["obs_mask"] = ~jnp.isnan(data_dict["observed_reference_value"])
        data_dict["data"] = None
        return data_dict

    def _predict(
        self, data: Dict, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray], optional
            A dictionary containing the predicted effects, by default None.

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries.
        """
        observed_reference_value = data["observed_reference_value"]
        obs_mask = data["obs_mask"]

        x = predicted_effects[self.effect_name]

        with numpyro.handlers.mask(mask=obs_mask):
            numpyro.sample(
                "exact_likelihood:ignore",
                dist.Normal(x, self.prior_scale),
                obs=observed_reference_value,
            )

        return x

    @classmethod
    def get_test_params(cls, parameter_set="default"):

        return [
            {
                "effect_name": "trend",
                "reference_df": pd.DataFrame({"y": [1, 2, 3]}),
                "prior_scale": 0.1,
            }
        ]



"""Definition of Linear Effect class."""

from typing import Any, Dict, Optional

import jax.numpy as jnp
import numpyro
from numpyro import distributions as dist
from numpyro.distributions import Distribution

from prophetverse.effects.base import (
    EFFECT_APPLICATION_TYPE,
    BaseAdditiveOrMultiplicativeEffect,
)
from prophetverse.utils.frame_to_array import series_to_tensor

__all__ = ["LinearEffect"]


class LinearEffect(BaseAdditiveOrMultiplicativeEffect):
    """Represents a linear effect in a hierarchical prophet model.

    Parameters
    ----------
    prior : Distribution, optional
        A numpyro distribution to use as prior. Defaults to dist.Normal(0, 1)
    effect_mode : effects_application, optional
        Either "multiplicative" or "additive" by default "multiplicative".
    """

    _tags = {
        "hierarchical_prophet_compliant": True,
        "capability:multivariate_input": True,
    }

    def __init__(
        self,
        effect_mode: EFFECT_APPLICATION_TYPE = "multiplicative",
        prior: Optional[Distribution] = None,
        broadcast=False,
    ):
        self.prior = prior
        self.broadcast = broadcast
        self._prior = self.prior if prior is not None else dist.Normal(0, 0.1)

        super().__init__(effect_mode=effect_mode)

        if self.broadcast:
            self.set_tags(**{"capability:multivariate_input": False})

    def _predict(
        self,
        data: Any,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """
        n_features = data.shape[-1]

        with numpyro.plate("features_plate", n_features, dim=-1):
            coefficients = numpyro.sample("coefs", self._prior)

        if coefficients.ndim == 1:
            coefficients = jnp.expand_dims(coefficients, axis=-1)

        if data.ndim == 3 and coefficients.ndim == 2:
            coefficients = jnp.expand_dims(coefficients, axis=0)

        return data @ coefficients


class PanelBHLinearEffect(BaseAdditiveOrMultiplicativeEffect):
    """
    Bayesian Hierarchical linear effect.

    Parameters
    ----------
    loc_hyperprior : Distribution, optional
        A numpyro distribution to use as the location hyperprior. Defaults to dist.Normal(0
        , 1).
    scale_hyperprior : Distribution, optional
        A numpyro distribution to use as the scale hyperprior. Defaults to dist.HalfNormal
        (1).
    prior_callable : Distribution, optional
        A numpyro distribution callable to use as the prior. Defaults to dist.Normal.
    effect_mode : effects_application, optional
        Either "multiplicative" or "additive" by default "multiplicative".
    """

    _tags = {
        "hierarchical_prophet_compliant": True,
        "capability:panel": True,
        "capability:multivariate_input": False,
        "feature:panel_hyperpriors": True,
    }

    def __init__(
        self,
        effect_mode: EFFECT_APPLICATION_TYPE = "multiplicative",
        loc_hyperprior: Optional[Distribution] = None,
        scale_hyperprior: Optional[Distribution] = None,
        prior_callable: Optional[Distribution] = None,
    ):
        self.loc_hyperprior = loc_hyperprior
        self.scale_hyperprior = scale_hyperprior
        self.prior_callable = prior_callable

        super().__init__(
            effect_mode=effect_mode,
        )

        self._loc_hyperprior = (
            loc_hyperprior if loc_hyperprior is not None else dist.Normal(0, 1)
        )
        self._scale_hyperprior = (
            scale_hyperprior if scale_hyperprior is not None else dist.HalfNormal(1)
        )
        self._prior_callable = (
            prior_callable if prior_callable is not None else dist.Normal
        )

    def _transform(self, X, fh):
        """Prepare input data to be passed to numpyro model."""
        return series_to_tensor(X)

    def _predict(
        self,
        data: Any,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """Apply and return the effect values.

        Parameters
        ----------
        data : Any
            Data obtained from the transformed method.

        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects

        Returns
        -------
        jnp.ndarray
            An array with shape (T,1) for univariate timeseries, or (N, T, 1) for
            multivariate timeseries, where T is the number of timepoints and N is the
            number of series.
        """

        loc_hyperprior = numpyro.sample(
            "loc_hyperprior",
            self._loc_hyperprior,
        )

        scale_hyperprior = numpyro.sample(
            "scale_hyperprior",
            self._scale_hyperprior,
        )

        args = (loc_hyperprior, scale_hyperprior)

        prior = self._prior_callable(*args)

        with numpyro.handlers.plate("panel_plate", data.shape[0], dim=-1):
            coefficients = numpyro.sample(
                "coefs",
                prior,
            )

        return data @ coefficients.reshape((-1, 1, 1))



"""Loaders for the datasets used in the examples."""

from pathlib import Path
from typing import Union

import pandas as pd
from sktime.datasets import load_forecastingdata
from sktime.transformations.hierarchical.aggregate import Aggregator

DATASET_MODULE_PATH = Path(__file__).parent


def load_tourism(groupby: Union[list[str], str] = "Region"):
    """
    Load the tourism dataset from Athanasopoulos et al. (2011).

    Parameters
    ----------
    groupby : list[str] or str, optional
        The columns to group by. Defaults to "Region".

    Returns
    -------
    pd.DataFrame
        The tourism dataset.

    See Also
    --------
    [1] Athanasopoulos, G., Hyndman, R. J., Song, H., & Wu, D. C. (2011).
    The tourism forecasting competition. International Journal of Forecasting,
    27(3), 822-844.
    """
    if isinstance(groupby, str):
        groupby = [groupby]

    groupby = [g.lower() for g in groupby]

    # Verify if the groupby columns are valid
    valid_columns = ["region", "purpose", "state", "quarter"]
    diff = set(groupby) - set(valid_columns)
    assert not diff, f"Invalid columns: {diff}"

    data = pd.read_csv(DATASET_MODULE_PATH / "raw/tourism.csv")
    data["Quarter"] = pd.PeriodIndex(data["Quarter"], freq="Q")
    data = data.set_index(["Region", "Purpose", "State", "Quarter"])[["Trips"]]
    data = data.sort_index()

    idxs_to_groupby = [g.capitalize() for g in groupby]

    y = (
        Aggregator(flatten_single_levels=False)
        .fit_transform(data)
        .groupby(level=[*idxs_to_groupby, -1])
        .sum()
    )

    return y


def load_peyton_manning():
    """Load the Peyton Manning dataset (used in Prophet's documentation).

    Returns
    -------
    pd.DataFrame
        the Peyton Manning dataset.
    """
    df = pd.read_csv(
        "https://raw.githubusercontent.com/facebook/prophet/main/"
        "examples/example_wp_log_peyton_manning.csv"
    )
    df["ds"] = pd.to_datetime(df["ds"]).dt.to_period("D")
    return df.set_index("ds")


def load_tensorflow_github_stars():
    """Load the TensorFlow GitHub stars dataset.

    Returns
    -------
    pd.DataFrame
        The TensorFlow GitHub stars dataset.
    """
    df = pd.read_csv(
        DATASET_MODULE_PATH / "raw/tensorflow_tensorflow-stars-history.csv"
    )
    df["date"] = pd.to_datetime(df["date"], format="%d-%m-%Y").dt.to_period("D")
    df = df.set_index("date")
    df = df.sort_index()
    y = df[["day-stars"]]

    return y


def load_pedestrian_count():
    """Load the pedestrian count dataset.

    Returns
    -------
    pd.DataFrame
        The pedestrian count dataset.
    """

    def _parse_data(df):

        dfs = []
        # iterrows
        for _, row in df.iterrows():

            _df = pd.DataFrame(
                data={
                    "pedestrian_count": row["series_value"],
                    "timestamp": pd.period_range(
                        row["start_timestamp"],
                        periods=len(row["series_value"]),
                        freq="H",
                    ),
                },
            )
            _df["series_name"] = row["series_name"]
            dfs.append(_df)

        return pd.concat(dfs).set_index(["series_name", "timestamp"])

    df, _ = load_forecastingdata("pedestrian_counts_dataset")
    return _parse_data(df)



"""Module containing the base class for trend models."""

import pandas as pd

from prophetverse.utils.frame_to_array import convert_index_to_days_since_epoch


class TrendEffectMixin:
    """
    Mixin class for trend models.

    Trend models are effects applied to the trend component of a time series.

    Attributes
    ----------
    t_scale: float
        The time scale of the trend model.
    t_start: float
        The starting time of the trend model.
    n_series: int
        The number of series in the time series data.
    """

    _tags = {
        "requires_X": False,
        "hierarchical_prophet_compliant": True,
        "capability:multivariate_input": True,
    }

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1) -> None:
        """Initialize the effect.

        Set the time scale, starting time, and number of series attributes.

        Parameters
        ----------
        y : pd.DataFrame
            The timeseries dataframe

        X : pd.DataFrame
            The DataFrame to initialize the effect.
        """
        # Set time scale
        t_days = convert_index_to_days_since_epoch(
            y.index.get_level_values(-1).unique()
        )
        self.t_scale = (t_days[1:] - t_days[:-1]).mean()
        self.t_start = t_days.min() / self.t_scale
        if y.index.nlevels > 1:
            self.n_series = y.index.droplevel(-1).nunique()
        else:
            self.n_series = 1

    def _index_to_scaled_timearray(self, idx):
        """
        Convert the index to a scaled time array.

        Parameters
        ----------
        idx: int
            The index to be converted.

        Returns
        -------
        float
            The scaled time array value.
        """
        if idx.nlevels > 1:
            idx = idx.get_level_values(-1).unique()

        t_days = convert_index_to_days_since_epoch(idx)
        return (t_days) / self.t_scale - self.t_start



"""Numpyro VI inference engines for prophet models.

The VIInferenceEngine class performs Variational Inference using SVI with
configurable autoguides specified as string parameters.
"""

from ast import Call
from typing import Callable, Optional, Union

import jax.numpy as jnp
import numpyro
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.autoguide import (
    AutoDiagonalNormal,
    AutoLowRankMultivariateNormal,
    AutoMultivariateNormal,
    AutoNormal,
)
from numpyro.infer.initialization import init_to_mean
from numpyro.infer.svi import SVIRunResult

from prophetverse.engine.base import BaseInferenceEngine
from prophetverse.engine.optimizer.optimizer import (
    BaseOptimizer,
    CosineScheduleAdamOptimizer,
)

_DEFAULT_PREDICT_NUM_SAMPLES = 1000
DEFAULT_PROGRESS_BAR = False

# Mapping of guide string names to numpyro autoguide classes
GUIDE_MAP = {
    "AutoNormal": AutoNormal,
    "AutoMultivariateNormal": AutoMultivariateNormal,
    "AutoDiagonalNormal": AutoDiagonalNormal,
    "AutoLowRankMultivariateNormal": AutoLowRankMultivariateNormal,
}


class VIInferenceEngine(BaseInferenceEngine):
    """
    Variational Inference Engine.

    This class performs Variational Inference using Stochastic Variational Inference
    (SVI) with configurable autoguides. It provides methods for inference and
    prediction.

    Parameters
    ----------
    guide : str or AutoGuide, optional
        The name of the autoguide to use for variational inference.
        Available options: "AutoNormal", "AutoMultivariateNormal",
        "AutoDiagonalNormal", "AutoLowRankMultivariateNormal".
        Default is "AutoNormal".
    optimizer : Optional[BaseOptimizer]
        The optimizer to be used for inference. If not provided, the default is
        AdamOptimizer.
    num_steps : int, optional
        The number of steps to run the optimizer. Default is 10000.
    num_samples : int, optional
        The number of samples to generate during prediction.
        Default is _DEFAULT_PREDICT_NUM_SAMPLES.
    rng_key : optional
        The random number generator key.
    progress_bar : bool, optional
        Whether to display a progress bar during inference. Default is
        DEFAULT_PROGRESS_BAR.
    stable_update : bool, optional
        Whether to use stable update during inference. Default is False.
    forward_mode_differentiation : bool, optional
        Whether to use forward mode differentiation. Default is False.
    init_scale : float, optional
        The scale for initializing the parameters. Default is 0.1.
    init_loc_fn : optional
        The function to initialize the location parameter. If not provided, the default
        is init_to_mean.

    """

    _tags = {
        "inference_method": "vi",
    }

    def __init__(
        self,
        guide: Optional[
            Union[str, Callable[[Callable], Callable]]
        ] = "AutoDiagonalNormal",
        optimizer: Optional[BaseOptimizer] = None,
        num_steps=10_000,
        num_samples=_DEFAULT_PREDICT_NUM_SAMPLES,
        rng_key=None,
        progress_bar: bool = DEFAULT_PROGRESS_BAR,
        stable_update=False,
        forward_mode_differentiation=False,
        init_scale=0.1,
        init_loc_fn=None,
    ):
        self.guide = guide
        self.optimizer = optimizer
        self.num_steps = num_steps
        self.num_samples = num_samples
        self.progress_bar = progress_bar
        self.stable_update = stable_update
        self.forward_mode_differentiation = forward_mode_differentiation
        self.init_scale = init_scale
        self.init_loc_fn = init_loc_fn
        super().__init__(rng_key)

        # Validate guide parameter
        if isinstance(guide, str) and guide not in GUIDE_MAP:
            available_guides = list(GUIDE_MAP.keys())
            raise ValueError(
                f"Unknown guide '{guide}'. Available guides: {available_guides}"
            )

        if optimizer is None:
            optimizer = CosineScheduleAdamOptimizer()

        self._optimizer = optimizer
        self._num_steps = num_steps

        if self._optimizer.get_tag("is_solver", False):  # type: ignore[union-attr]
            # If solver, there's a single "solver step". For compatibility,
            # we set num_steps to 1 and max_iter to the original num_steps.

            self._optimizer = self._optimizer.set_max_iter(  # type: ignore[union-attr]
                self._num_steps
            )
            self._num_steps = 1

    def _build_guide(self, model: Callable) -> Callable:
        """
        Build the guide function based on the specified guide type.

        Returns
        -------
        Callable[[...], None]
            The constructed guide function.
        """
        if isinstance(self.guide, str):
            guide_class = GUIDE_MAP[self.guide]
            return guide_class(
                model,
                init_loc_fn=self.init_loc_fn or init_to_mean(),
                init_scale=self.init_scale,
            )
        elif callable(self.guide):
            return self.guide(model)
        else:
            raise ValueError("Invalid guide type. Must be a string or callable.")

    def _infer(self, **kwargs):
        """
        Perform Variational Inference.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        self
            The updated VIInferenceEngine object.
        """
        self.guide_ = self._build_guide(self.model_)

        def get_result(
            rng_key,
            model,
            guide,
            optimizer,
            num_steps,
            progress_bar,
            stable_update,
            forward_mode_differentiation,
            **kwargs,
        ) -> SVIRunResult:
            svi_ = SVI(
                model,
                guide,
                optimizer,
                loss=Trace_ELBO(),
            )
            return svi_.run(
                rng_key=rng_key,
                progress_bar=progress_bar,
                stable_update=stable_update,
                num_steps=num_steps,
                forward_mode_differentiation=forward_mode_differentiation,
                **kwargs,
            )

        self.run_results_: SVIRunResult = get_result(
            self._rng_key,
            self.model_,
            self.guide_,
            self._optimizer.create_optimizer(),
            self._num_steps,
            stable_update=self.stable_update,
            progress_bar=self.progress_bar,
            forward_mode_differentiation=self.forward_mode_differentiation,
            **kwargs,
        )

        self.raise_error_if_nan_loss(self.run_results_)

        self.posterior_samples_ = self.guide_.sample_posterior(
            self._rng_key, params=self.run_results_.params, **kwargs
        )
        return self

    def raise_error_if_nan_loss(self, run_results: SVIRunResult):
        """
        Raise an error if the loss is NaN.

        Parameters
        ----------
        run_results : SVIRunResult
            The result of the SVI run.

        Raises
        ------
        VIInferenceEngineError
            If the last loss is NaN.
        """
        losses = run_results.losses
        if jnp.isnan(losses)[-1]:
            msg = "NaN losses in VIInferenceEngine."
            msg += " Try decreasing the learning rate or changing the model specs."
            msg += " If the problem persists, please open an issue at"
            msg += " https://github.com/felipeangelimvieira/prophetverse"
            raise VIInferenceEngineError(msg)

    def _predict(self, **kwargs):
        """
        Generate predictions using the trained model.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to be passed to the model.

        Returns
        -------
        dict
            The predicted samples generated by the model.
        """
        predictive = numpyro.infer.Predictive(
            self.model_,
            params=self.run_results_.params,
            guide=self.guide_,
            num_samples=self.num_samples,
        )
        self.samples_ = predictive(rng_key=self._rng_key, **kwargs)
        return self.samples_

    @classmethod
    def get_test_params(*args, **kwargs):
        """Return test params for unit testing."""
        return [
            {
                "guide": "AutoNormal",
                "optimizer": CosineScheduleAdamOptimizer(),
                "num_steps": 100,
            },
            {
                "guide": "AutoMultivariateNormal",
                "optimizer": CosineScheduleAdamOptimizer(),
                "num_steps": 100,
            },
            {
                "guide": "AutoDiagonalNormal",
                "optimizer": CosineScheduleAdamOptimizer(),
                "num_steps": 100,
            },
        ]


class VIInferenceEngineError(Exception):
    """Exception raised for NaN losses in VIInferenceEngine."""

    def __init__(self, message="NaN losses in VIInferenceEngine"):
        self.message = message
        super().__init__(self.message)



"""Synthetic datasets for testing and examples."""

from ._composite_effect_example import load_composite_effect_example
from ._squared_exogenous import load_synthetic_squared_exogenous

__all__ = ["load_composite_effect_example", "load_synthetic_squared_exogenous"]



"""Truncated discrete distribution."""

import jax
import jax.numpy as jnp
import numpyro.distributions as dist
from jax import random
from numpyro.distributions import constraints
from numpyro.distributions.distribution import Distribution
from numpyro.distributions.util import validate_sample

from ._inverse_functions import _inverse_neg_binom, _inverse_poisson

REGISTRY = {
    dist.Poisson: _inverse_poisson,
    dist.NegativeBinomial2: _inverse_neg_binom,
}


class TruncatedDiscrete(Distribution):
    """Wrapper to create a zero-truncated version of a discrete distribution.

    Takes a base distribution (like Poisson or NegativeBinomial) and
    modifies its log_prob and sample methods to enforce a support of {1, 2, 3, ...}.

    Parameters
    ----------
    base_dist: numpyro.distributions.Distribution
        A discrete distribution with non-negative support.

    low: int, default=0
        The lower bound of the support. The distribution will only consider
        values greater than this bound.
    """

    has_rsample = False
    pytree_data_fields = ("base_dist",)
    pytree_aux_fields = ("_icdf", "low")

    def __init__(self, base_dist, low: int = 0, *, validate_args=None):
        if base_dist.support is not constraints.nonnegative_integer:
            raise ValueError("ZeroTruncated only works with discrete distributions!")

        if base_dist.__class__ not in REGISTRY:
            raise ValueError(
                f"Base distribution '{base_dist.__class__.__name__}' not supported!"
            )

        super().__init__(
            batch_shape=base_dist.batch_shape,
            event_shape=base_dist.event_shape,
            validate_args=validate_args,
        )

        self.base_dist = base_dist
        self.low = low
        self._icdf = REGISTRY[base_dist.__class__]

    @property
    def support(self):  # noqa: D102
        return constraints.integer_greater_than(self.low)

    @validate_sample
    def log_prob(self, value):  # noqa: D102
        is_invalid = value <= self.low

        log_prob_base = self.base_dist.log_prob(value)

        log_prob_at_zero = self.base_dist.log_prob(self.low)
        log_normalizer = jnp.log1p(-jnp.exp(log_prob_at_zero))

        log_prob_truncated = log_prob_base - log_normalizer

        return jnp.where(is_invalid, -jnp.inf, log_prob_truncated)

    def sample(self, key, sample_shape=()):  # noqa: D102
        shape = sample_shape + self.batch_shape

        dtype = jnp.result_type(float)
        finfo = jnp.finfo(dtype)
        minval = finfo.tiny

        u = random.uniform(key, shape=shape, minval=minval)

        return self.icdf(u)

    def icdf(self, u):  # noqa: D102
        result_shape = jax.ShapeDtypeStruct(u.shape, jnp.result_type(float))

        low_cdf = self.base_dist.cdf(self.low)
        normalizer = 1.0 - low_cdf
        x = normalizer * u + low_cdf

        result = jax.pure_callback(
            self._icdf,
            result_shape,
            *(self.base_dist, x),
        )
        return result.astype(jnp.result_type(int))



"""Simulate data from a model, and intervene optionally."""

from typing import Dict, Optional, Union

import jax.numpy as jnp
import numpy as np
import numpyro
import pandas as pd
from jax.random import PRNGKey

from prophetverse.sktime.base import BaseProphetForecaster
from prophetverse._model import model as model_func
from prophetverse.effects.target.base import BaseTargetEffect

from numpyro.primitives import Messenger


class IgnoreObservedSites(Messenger):
    def __init__(self):
        """
        obs_map: a dict mapping site names -> values
        """
        super().__init__(fn=None)

    def process_message(self, msg):
        # only intercept real sample sites that have no value yet

        if msg["type"] == "sample":
            # ...but tell NumPyro it's NOT an observed site
            msg["is_observed"] = False
            msg["obs"] = None


def simulate(
    model: BaseProphetForecaster,
    fh: pd.Index,
    X: Optional[pd.DataFrame] = None,
    y: Optional[pd.DataFrame] = None,
    do: Optional[Dict[str, Union[jnp.ndarray, float]]] = None,
    return_model=False,
):
    """
    Simulate data from a model.

    **EXPERIMENTAL FEATURE**
    This feature allow to do prior predictive checks and to intervene to
    obtain simulated data.

    Parameters
    ----------
    model : BaseProphetForecaster
        The probabilistic model to perform inference on.
    fh : pd.Index
        The forecasting horizon as a pandas Index.
    X : pd.DataFrame, optional
        The input DataFrame containing the exogenous variables.
    y : pd.DataFrame, optional
        The timeseries dataframe. This is used by effects that implement `_fit` and
        use the target timeseries to initialize some parameters. If not provided,
        a dummy y will be created.
    do : Dict, optional
        A dictionary with the variables to intervene and their values.
    num_samples : int, optional
        The number of samples to generate. Defaults to 10.
    return_model : bool, optional
        If True, the fitted model will be returned. Defaults to False.
    Returns
    -------
    Union[Dict, Tuple]
        If return_model=False, a dictionary with the simulated data. Otherwise,
        a tuple (simulated_data, model).
    """
    # Fit model, creating a dummy y if it is not provided
    if y is None:
        if X is None:
            index = fh
        else:
            index, _ = X.index.reindex(fh, level=-1)
        y = pd.DataFrame(
            index=index, data=np.random.rand(len(index)) * 10, columns=["dummy"]
        )

    model = model.clone()
    model.fit(X=X, y=y)

    with IgnoreObservedSites():
        if do is not None:
            with numpyro.handlers.do(data=do):
                components = model.predict_component_samples(X=X, fh=fh)
        else:
            components = model.predict_component_samples(X=X, fh=fh)
    if return_model:
        return components, model
    return components



"""Definition of Chained Effects class."""

from typing import Any, Dict, List, Tuple

import jax.numpy as jnp
from skbase.base import BaseMetaEstimatorMixin
import numpyro
from prophetverse.effects.base import BaseEffect

__all__ = ["ChainedEffects"]


class ChainedEffects(BaseMetaEstimatorMixin, BaseEffect):
    """
    Chains multiple effects sequentially, applying them one after the other.

    Parameters
    ----------
    steps : List[BaseEffect]
        A list of effects to be applied sequentially.
    """

    _tags = {
        "hierarchical_prophet_compliant": True,
        "requires_X": True,
        "filter_indexes_with_forecating_horizon_at_transform": True,
        "named_object_parameters": "named_steps",
    }

    def __init__(self, steps: List[Tuple[str, BaseEffect]]):
        self.steps = steps
        super().__init__()

        self.named_steps = []
        for i, val in enumerate(self.steps):
            if isinstance(val, tuple):
                self.named_steps.append(val)
            elif isinstance(val, BaseEffect):
                self.named_steps.append((str(i), val))
            else:
                raise ValueError(
                    f"Invalid type {type(val)} for step {i}. Must be a tuple or BaseEffect."
                )

        self.set_tags(**{"requires_X": steps[0][1].get_tag("requires_X", False)})

    def _fit(self, y: Any, X: Any, scale: float = 1.0):
        """
        Fit all chained effects sequentially.

        Parameters
        ----------
        y : Any
            Target data (e.g., time series values).
        X : Any
            Exogenous variables.
        scale : float, optional
            Scale of the timeseries.
        """
        self.named_steps_ = []
        for name, effect in self.named_steps:
            effect = effect.clone()
            effect.fit(y, X, scale)
            self.named_steps_.append((name, effect))

    def _transform(self, X: Any, fh: Any) -> Any:
        """
        Transform input data sequentially through all chained effects.

        Parameters
        ----------
        X : Any
            Input data (e.g., exogenous variables).
        fh : Any
            Forecasting horizon.

        Returns
        -------
        Any
            Transformed data after applying all effects.
        """
        output = X
        output = self.named_steps_[0][1].transform(output, fh)
        return output

    def _predict(
        self,
        data: jnp.ndarray,
        predicted_effects: Dict[str, jnp.ndarray],
        *args,
        **kwargs,
    ) -> jnp.ndarray:
        """
        Apply all chained effects sequentially.

        Parameters
        ----------
        data : jnp.ndarray
            Data obtained from the transformed method (shape: T, 1).
        predicted_effects : Dict[str, jnp.ndarray]
            A dictionary containing the predicted effects.
        params : Dict[str, Dict[str, jnp.ndarray]]
            A dictionary containing the sampled parameters for each effect.

        Returns
        -------
        jnp.ndarray
            The transformed data after applying all effects.
        """
        output = data
        for name, effect in self.named_steps_:
            with numpyro.handlers.scope(prefix=name):
                output = effect.predict(output, predicted_effects)
        return output

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        from prophetverse.effects.linear import LinearEffect
        from prophetverse.effects.adstock import GeometricAdstockEffect

        return [
            {
                "steps": [
                    ("adstock", GeometricAdstockEffect()),
                    ("linear", LinearEffect()),
                ]
            },
            {
                "steps": [
                    ("linear", LinearEffect()),
                    ("adstock", GeometricAdstockEffect()),
                ]
            },
        ]



"""Hurdle distribution implementation."""

import jax
import jax.numpy as jnp
from numpyro.distributions import Bernoulli, Distribution, constraints
from numpyro.distributions.util import validate_sample

from prophetverse.distributions.truncated_discrete import TruncatedDiscrete


class HurdleDistribution(Distribution):
    """A Hurdle distribution.

    This distribution models data with an excess of zeros. It is a mixture of a
    point mass at zero and a positive distribution for values greater than zero.
    This implementation creates the positive distribution by internally truncating
    a given base distribution at `low=0`.

    Parameters
    ----------
    prob_gt_zero: jnp.ndarray
        Probability of observing a value greater than zero.

    positive_dist: TruncatedDiscrete
        Truncated distribution to use for positive values.
    """

    arg_constraints = {"prob_gt_zero": constraints.unit_interval}
    support = constraints.nonnegative
    has_rsample = False

    pytree_data_fields = ("prob_gt_zero", "positive_dist")

    def __init__(
        self,
        prob_gt_zero: jnp.ndarray,
        positive_dist: TruncatedDiscrete,
        validate_args=None,
    ):
        self.prob_gt_zero = prob_gt_zero
        self.positive_dist = positive_dist

        batch_shape = jnp.broadcast_shapes(
            jnp.shape(prob_gt_zero), self.positive_dist.batch_shape
        )

        super().__init__(
            batch_shape=batch_shape,
            event_shape=self.positive_dist.event_shape,
            validate_args=validate_args,
        )

    @validate_sample
    def log_prob(self, value):  # noqa: D102
        log_prob_zero = jnp.log1p(-self.prob_gt_zero)
        log_prob_positive = jnp.log(self.prob_gt_zero) + self.positive_dist.log_prob(
            value
        )

        is_zero = value == 0
        return jnp.where(is_zero, log_prob_zero, log_prob_positive)

    def sample(self, key, sample_shape=()):  # noqa: D102
        key_hurdle, key_positive = jax.random.split(key)
        is_positive = Bernoulli(probs=self.prob_gt_zero).sample(
            key_hurdle, sample_shape
        )

        positive_values = self.positive_dist.sample(key_positive, sample_shape)

        return jnp.where(is_positive, positive_values, 0)



"""Coupled effects"""

from typing import Any, Dict

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pandas as pd

from prophetverse.utils.frame_to_array import series_to_tensor_or_array

from .base import BaseEffect

__all__ = ["CoupledExactLikelihood"]


class CoupledExactLikelihood(BaseEffect):
    """Link two effects via a Normal likelihood"""

    _tags = {"requires_X": False, "hierarchical_prophet_compliant": False}

    def __init__(
        self,
        source_effect_name: str,
        target_effect_name: str,
        prior_scale: float,
    ):
        self.source_effect_name = source_effect_name
        self.target_effect_name = target_effect_name
        self.prior_scale = prior_scale
        assert prior_scale > 0, "prior_scale must be greater than 0"
        super().__init__()

    def _fit(self, y: pd.DataFrame, X: pd.DataFrame, scale: float = 1):
        self.timeseries_scale = scale

    def _transform(self, X: pd.DataFrame, fh: pd.Index) -> Dict[str, Any]:
        return {"data": None}

    def _predict(
        self, data: Dict, predicted_effects: Dict[str, jnp.ndarray], *args, **kwargs
    ) -> jnp.ndarray:
        source = predicted_effects[self.source_effect_name]
        target = predicted_effects[self.target_effect_name]
        numpyro.sample(
            f"coupled_exact_likelihood:{self.target_effect_name}",
            dist.Normal(source, self.prior_scale),
            obs=target,
        )
        return target

    @classmethod
    def get_test_params(cls, parameter_set: str = "default"):
        return [
            {
                "source_effect_name": "trend",
                "target_effect_name": "trend",
                "prior_scale": 0.1,
            }
        ]



"""Utilities module."""

from .frame_to_array import (
    convert_dataframe_to_tensors,
    convert_index_to_days_since_epoch,
    iterate_all_series,
    series_to_tensor,
    series_to_tensor_or_array,
)
from .multiindex import (
    get_bottom_series_idx,
    get_multiindex_loc,
    loc_bottom_series,
    reindex_time_series,
)
from .regex import exact, no_input_columns, starts_with

__all__ = [
    "get_bottom_series_idx",
    "get_multiindex_loc",
    "loc_bottom_series",
    "iterate_all_series",
    "convert_index_to_days_since_epoch",
    "series_to_tensor",
    "series_to_tensor_or_array",
    "convert_dataframe_to_tensors",
    "reindex_time_series",
    "exact",
    "starts_with",
    "no_input_columns",
]
